# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from typing import Any, List, Tuple, Union

import paddle
import paddle.nn as nn
import paddle.nn.functional as F
from paddle import ParamAttr
from paddle.nn.initializer import Constant


def get_bias_attr(k: float) -> ParamAttr:
    """
    get_bias_attr

    Args:
        k (float): Scaling factor for standard deviation calculation

    Returns:
        ParamAttr: Parameter attribute with uniform initializer
    """
    stdv = 1.0 / math.sqrt(k * 1.0)
    initializer = paddle.nn.initializer.Uniform(-stdv, stdv)
    bias_attr = ParamAttr(initializer=initializer)
    return bias_attr


class LearnableAffineBlock(nn.Layer):
    """
    LearnableAffineBlock

    Args:
        scale_value (float, optional): Initial value for scale parameter, default is 1.0
        bias_value (float, optional): Initial value for bias parameter, default is 0.0
        lr_mult (float, optional): Learning rate multiplier for base learning rate, default is 1.0
        lab_lr (float, optional): Additional learning rate multiplier for affine parameters, default is 0.01

    Returns:
        paddle.Tensor: Output tensor after affine transformation (scale * x + bias)
    """

    def __init__(
        self,
        scale_value: float = 1.0,
        bias_value: float = 0.0,
        lr_mult: float = 1.0,
        lab_lr: float = 0.01,
    ) -> None:
        super().__init__()
        self.scale = self.create_parameter(
            shape=[
                1,
            ],
            default_initializer=Constant(value=scale_value),
            attr=ParamAttr(learning_rate=lr_mult * lab_lr),
        )
        self.add_parameter("scale", self.scale)
        self.bias = self.create_parameter(
            shape=[
                1,
            ],
            default_initializer=Constant(value=bias_value),
            attr=ParamAttr(learning_rate=lr_mult * lab_lr),
        )
        self.add_parameter("bias", self.bias)

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        return self.scale * x + self.bias


class Head(nn.Layer):
    """
    Head

    Args:
        in_channels (int): Number of input channels
        kernel_list (List[int]): List of kernel sizes for conv/transposed conv layers
        fix_nan (bool): Whether to fix NaN issues (unused in current implementation)
        **kwargs: Additional keyword arguments

    Returns:
        paddle.Tensor: 1-channel sigmoid output tensor (or tuple with feature tensor if return_f=True)
    """

    def __init__(
        self, in_channels: int, kernel_list: List[int], fix_nan: bool, **kwargs: Any
    ) -> None:
        super(Head, self).__init__()

        self.conv1 = nn.Conv2D(
            in_channels=in_channels,
            out_channels=in_channels // 4,
            kernel_size=kernel_list[0],
            padding=int(kernel_list[0] // 2),
            weight_attr=ParamAttr(),
            bias_attr=False,
        )
        self.conv_bn1 = nn.BatchNorm(
            num_channels=in_channels // 4,
            param_attr=ParamAttr(initializer=paddle.nn.initializer.Constant(value=1.0)),
            bias_attr=ParamAttr(initializer=paddle.nn.initializer.Constant(value=1e-4)),
            act="relu",
        )

        self.conv2 = nn.Conv2DTranspose(
            in_channels=in_channels // 4,
            out_channels=in_channels // 4,
            kernel_size=kernel_list[1],
            stride=2,
            weight_attr=ParamAttr(initializer=paddle.nn.initializer.KaimingUniform()),
            bias_attr=get_bias_attr(in_channels // 4),
        )
        self.conv_bn2 = nn.BatchNorm(
            num_channels=in_channels // 4,
            param_attr=ParamAttr(initializer=paddle.nn.initializer.Constant(value=1.0)),
            bias_attr=ParamAttr(initializer=paddle.nn.initializer.Constant(value=1e-4)),
            act="relu",
        )
        self.conv3 = nn.Conv2DTranspose(
            in_channels=in_channels // 4,
            out_channels=1,
            kernel_size=kernel_list[2],
            stride=2,
            weight_attr=ParamAttr(initializer=paddle.nn.initializer.KaimingUniform()),
            bias_attr=get_bias_attr(in_channels // 4),
        )

        self.fix_nan = fix_nan

    def forward(
        self, x: paddle.Tensor, return_f: bool = False
    ) -> Union[paddle.Tensor, Tuple]:
        x = self.conv1(x)
        x = self.conv_bn1(x)
        x = self.conv2(x)
        x = self.conv_bn2(x)
        if return_f is True:
            f = x
        x = self.conv3(x)
        x = F.sigmoid(x)
        if return_f is True:
            return x, f
        return x


class DBHead(nn.Layer):
    """
    DBHead

    Paper: https://arxiv.org/abs/1911.08947

    Args:
        in_channels (int): Number of input channels
        k (int): DB head hyperparameter (kernel factor)
        **kwargs: Additional keyword arguments for Head class (kernel_list, fix_nan)

    Returns:
        paddle.Tensor: Shrinkage map tensor after DB binarization
    """

    def __init__(self, in_channels: int, k: int, **kwargs) -> None:
        super(DBHead, self).__init__()
        self.k = k
        self.binarize = Head(in_channels, **kwargs)

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        shrink_maps = self.binarize(x)
        return shrink_maps



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, List, Optional

import paddle
import paddle.nn as nn
import paddle.nn.functional as F

from ....utils.benchmark import add_inference_operations, benchmark
from ...common.transformers.transformers import (
    BatchNormHFStateDictMixin,
    PretrainedModel,
)
from ._config import UVDocNetConfig


def conv3x3(
    in_channels: int, out_channels: int, kernel_size: int, stride: int = 1
) -> nn.Conv2D:
    """
    conv3x3: 3x3 convolution layer with same padding

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int): Kernel size of convolution layer
        stride (int, optional): Convolution stride, default is 1

    Returns:
        nn.Conv2D: Convolutional layer with same padding (padding = kernel_size // 2)
    """

    return nn.Conv2D(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=kernel_size,
        stride=stride,
        padding=kernel_size // 2,
    )


def dilated_conv_bn_act(
    in_channels: int, out_channels: int, dilation: int
) -> nn.Sequential:
    """
    dilated_conv_bn_act: Dilated convolution block with BN and ReLU activation

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        dilation (int): Dilation rate for dilated convolution

    Returns:
        nn.Sequential: Sequential block containing dilated conv, BN and ReLU layers
    """

    model = nn.Sequential(
        nn.Conv2D(
            in_channels=in_channels,
            out_channels=out_channels,
            bias_attr=False,
            kernel_size=3,
            stride=1,
            padding=dilation,
            dilation=dilation,
        ),
        nn.BatchNorm2D(out_channels),
        nn.ReLU(),
    )
    return model


def dilated_conv(
    in_channels: int,
    out_channels: int,
    kernel_size: int,
    dilation: int,
    stride: int = 1,
) -> nn.Sequential:
    """
    dilated_conv: Pure dilated convolution layer for multi-scale feature learning

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int): Kernel size of dilated convolution
        dilation (int): Dilation rate for dilated convolution
        stride (int, optional): Convolution stride, default is 1

    Returns:
        nn.Sequential: Sequential block containing only the dilated convolution layer
    """
    model = nn.Sequential(
        nn.Conv2D(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=dilation * (kernel_size // 2),
            dilation=dilation,
        )
    )
    return model


class ResidualBlockWithDilation(nn.Layer):
    """
    ResidualBlockWithDilation: Residual block with optional dilated convolution and downsampling

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int): Kernel size of convolution layers
        stride (int, optional): Convolution stride for first conv layer, default is 1
        downsample (Optional[nn.Layer]): Downsampling layer for residual connection, default is None
        is_activation (bool, optional): Whether to apply activation (unused in current implementation), default is True
        is_top (bool, optional): Whether it is the top block (uses standard conv instead of dilated conv), default is False

    Returns:
        paddle.Tensor: Output tensor after residual block with conv/dilated conv and ReLU activation
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        downsample: Optional[nn.Layer] = None,
        is_activation: bool = True,
        is_top: bool = False,
    ):
        super(ResidualBlockWithDilation, self).__init__()
        self.stride = stride
        self.downsample = downsample
        self.is_activation = is_activation
        self.is_top = is_top
        if self.stride != 1 or self.is_top:
            self.conv1 = conv3x3(in_channels, out_channels, kernel_size, self.stride)
            self.conv2 = conv3x3(out_channels, out_channels, kernel_size)
        else:
            self.conv1 = dilated_conv(
                in_channels, out_channels, kernel_size, dilation=3
            )
            self.conv2 = dilated_conv(
                out_channels, out_channels, kernel_size, dilation=3
            )
        self.bn1 = nn.BatchNorm2D(out_channels)
        self.relu = nn.ReLU()
        self.bn2 = nn.BatchNorm2D(out_channels)

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        residual = x
        if self.downsample is not None:
            residual = self.downsample(x)
        out1 = self.relu(self.bn1(self.conv1(x)))
        out2 = self.bn2(self.conv2(out1))
        out2 += residual
        out = self.relu(out2)
        return out


class ResnetStraight(nn.Layer):
    """
    ResnetStraight: Straightforward ResNet architecture with residual blocks and optional dilated convolution

    Args:
        num_filter (int): Base number of filters/channels for the network
        map_num (List[int]): List of channel scaling factors for each layer
        block_nums (List[int]): List of residual block numbers for each layer
        kernel_size (int): Kernel size of convolution layers in residual blocks
        stride (List[int]): List of stride values for each layer's first residual block

    Returns:
        paddle.Tensor: Output tensor from the third residual layer of the ResNet
    """

    def __init__(
        self,
        num_filter: int,
        map_num: List[int],
        block_nums: List[int],
        kernel_size: int,
        stride: List[int],
    ):
        super(ResnetStraight, self).__init__()
        self.in_channels = num_filter * map_num[0]
        self.stride = stride
        self.block_nums = block_nums
        self.kernel_size = kernel_size

        for layer_idx, (map_num_val, block_num, stride_val) in enumerate(
            zip(map_num[:3], block_nums[:3], stride[:3])
        ):
            layer = self.blocklayer(
                num_filter * map_num_val,
                block_num,
                kernel_size=self.kernel_size,
                stride=stride_val,
            )
            setattr(self, f"layer{layer_idx + 1}", layer)

    def blocklayer(
        self, out_channels: int, block_nums: int, kernel_size: int, stride: int = 1
    ) -> nn.Sequential:
        downsample = None
        if stride != 1 or self.in_channels != out_channels:
            downsample = nn.Sequential(
                conv3x3(
                    self.in_channels,
                    out_channels,
                    kernel_size=kernel_size,
                    stride=stride,
                ),
                nn.BatchNorm2D(out_channels),
            )

        layers = []
        for i in range(block_nums):
            layers.append(
                ResidualBlockWithDilation(
                    in_channels=self.in_channels if i == 0 else out_channels,
                    out_channels=out_channels,
                    kernel_size=kernel_size,
                    stride=stride if i == 0 else 1,
                    downsample=downsample if i == 0 else None,
                    is_top=True if i == 0 else False,
                )
            )
        self.in_channels = out_channels
        return nn.Sequential(*layers)

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        out1 = self.layer1(x)
        out2 = self.layer2(out1)
        out3 = self.layer3(out2)
        return out3


class UVDocNet(BatchNormHFStateDictMixin, PretrainedModel):
    """
    UVDocNet: Image rectification network based on ResNet and multi-scale dilated convolution

    Args:
        config (UVDocNetConfig): Configuration object containing network hyperparameters

    Returns:
        List: List containing the transformed document image tensor (converted to numpy array on CPU)
    """

    config_class = UVDocNetConfig

    def __init__(self, config: UVDocNetConfig):
        super(UVDocNet, self).__init__(config)

        self.num_filter = config.num_filter
        self.in_channels = config.in_channels
        self.kernel_size = config.kernel_size
        self.stride = config.stride
        self.map_num = config.map_num
        self.block_nums = config.block_nums
        self.dilation_values = config.dilation_values
        self.padding_mode = config.padding_mode
        self.upsample_size = config.upsample_size
        self.upsample_mode = config.upsample_mode

        self.resnet_head = nn.Sequential(
            nn.Conv2D(
                in_channels=self.in_channels,
                out_channels=self.num_filter * self.map_num[0],
                bias_attr=False,
                kernel_size=self.kernel_size,
                stride=2,
                padding=self.kernel_size // 2,
            ),
            nn.BatchNorm2D(self.num_filter * self.map_num[0]),
            nn.ReLU(),
            nn.Conv2D(
                in_channels=self.num_filter * self.map_num[0],
                out_channels=self.num_filter * self.map_num[0],
                bias_attr=False,
                kernel_size=self.kernel_size,
                stride=2,
                padding=self.kernel_size // 2,
            ),
            nn.BatchNorm2D(self.num_filter * self.map_num[0]),
            nn.ReLU(),
        )

        self.resnet_down = ResnetStraight(
            self.num_filter,
            self.map_num,
            block_nums=self.block_nums,
            kernel_size=self.kernel_size,
            stride=self.stride,
        )

        bridge_in_channels = self.num_filter * self.map_num[2]

        def _build_bridge(bridge_key: str) -> nn.Sequential:
            dilation = self.dilation_values[bridge_key]
            if isinstance(dilation, int):
                return nn.Sequential(
                    dilated_conv_bn_act(
                        bridge_in_channels, bridge_in_channels, dilation=dilation
                    )
                )
            else:
                return nn.Sequential(
                    *[
                        dilated_conv_bn_act(
                            bridge_in_channels, bridge_in_channels, dilation=d
                        )
                        for d in dilation
                    ]
                )

        self.bridge_1 = _build_bridge("bridge_1")
        self.bridge_2 = _build_bridge("bridge_2")
        self.bridge_3 = _build_bridge("bridge_3")
        self.bridge_4 = _build_bridge("bridge_4")
        self.bridge_5 = _build_bridge("bridge_5")
        self.bridge_6 = _build_bridge("bridge_6")

        self.bridge_concat = nn.Sequential(
            nn.Conv2D(
                in_channels=self.num_filter * self.map_num[2] * 6,
                out_channels=self.num_filter * self.map_num[2],
                bias_attr=False,
                kernel_size=1,
                stride=1,
                padding=0,
            ),
            nn.BatchNorm2D(self.num_filter * self.map_num[2]),
            nn.ReLU(),
        )

        self.out_point_positions2D = nn.Sequential(
            nn.Conv2D(
                in_channels=self.num_filter * self.map_num[2],
                out_channels=self.num_filter * self.map_num[0],
                bias_attr=False,
                kernel_size=self.kernel_size,
                stride=1,
                padding=self.kernel_size // 2,
                padding_mode=self.padding_mode,
            ),
            nn.BatchNorm2D(self.num_filter * self.map_num[0]),
            nn.PReLU(),
            nn.Conv2D(
                in_channels=self.num_filter * self.map_num[0],
                out_channels=2,
                kernel_size=self.kernel_size,
                stride=1,
                padding=self.kernel_size // 2,
                padding_mode=self.padding_mode,
            ),
        )

    add_inference_operations("uvdoc_forward")

    @benchmark.timeit_with_options(name="uvdoc_forward")
    def forward(self, x: Any) -> List[paddle.Tensor]:
        x = paddle.to_tensor(x[0])

        image = x
        h_ori, w_ori = x.shape[2:]
        x = F.upsample(
            x,
            size=(self.upsample_size[0], self.upsample_size[1]),
            mode=self.upsample_mode,
            align_corners=True,
        )
        resnet_head = self.resnet_head(x)
        resnet_down = self.resnet_down(resnet_head)

        bridge_1 = self.bridge_1(resnet_down)
        bridge_2 = self.bridge_2(resnet_down)
        bridge_3 = self.bridge_3(resnet_down)
        bridge_4 = self.bridge_4(resnet_down)
        bridge_5 = self.bridge_5(resnet_down)
        bridge_6 = self.bridge_6(resnet_down)

        bridge_concat = paddle.concat(
            x=[bridge_1, bridge_2, bridge_3, bridge_4, bridge_5, bridge_6], axis=1
        )
        bridge = self.bridge_concat(bridge_concat)
        out_point_positions2D = self.out_point_positions2D(bridge)

        bm_up = F.upsample(
            out_point_positions2D,
            size=(h_ori, w_ori),
            mode=self.upsample_mode,
            align_corners=True,
        )
        bm = bm_up.transpose([0, 2, 3, 1])
        out = F.grid_sample(image, bm, align_corners=True)

        return [out.cpu().numpy()]

    def _get_forward_key_rules(self):
        default_rules = super()._get_forward_key_rules()
        custom_rules = [("out_point_positions2D.2._weight", "_weight", "weight")]
        return default_rules + custom_rules

    def _get_reverse_key_rules(self):
        default_rules = super()._get_reverse_key_rules()
        custom_rules = [("out_point_positions2D.2.weight", "weight", "_weight")]
        return default_rules + custom_rules



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import random
from functools import partial

import numpy as np
from PIL import Image, ImageDraw

from ....utils import logging
from ....utils.deps import class_requires_deps, is_dep_available
from ....utils.fonts import SIMFANG_FONT
from ...common.result import (
    BaseCVResult,
    BaseResult,
    HtmlMixin,
    JsonMixin,
    MarkdownMixin,
    XlsxMixin,
)
from ..layout_parsing.result_v2 import (
    format_centered_by_html,
    format_first_line_func,
    format_image_plain_func,
    format_image_scaled_by_html_func,
    format_para_title_func,
    format_text_plain_func,
    format_title_func,
    simplify_table_func,
)
from ..ocr.result import draw_box_txt_fine, get_minarea_rect

SKIP_ORDER_LABELS = [
    "figure_title",
    "vision_footnote",
    "image",
    "chart",
    "table",
    "header",
    "header_image",
    "footer",
    "footer_image",
    "footnote",
    "aside_text",
]

if is_dep_available("opencv-contrib-python"):
    import cv2


class PaddleOCRVLBlock(object):
    """PaddleOCRVL Block Class"""

    def __init__(
        self,
        label,
        bbox,
        content="",
        group_id=None,
        polygon_points=None,
        global_block_id=None,
        global_group_id=None,
    ) -> None:
        """
        Initialize a PaddleOCRVLBlock object.

        Args:
            label (str): Label assigned to the block.
            bbox (list): Bounding box coordinates of the block.
            content (str, optional): Content of the block. Defaults to an empty string.
        """
        self.label = label
        self.bbox = list(map(int, bbox))
        self.content = content
        self.image = None
        self.polygon_points = polygon_points
        self.group_id = group_id
        self.global_block_id = global_block_id
        self.global_group_id = global_group_id

    def __str__(self) -> str:
        """
        Return a string representation of the block.
        """
        _str = f"\n\n#################\nlabel:\t{self.label}\nbbox:\t{self.bbox}\ncontent:\t{self.content}\n#################"
        return _str

    def __repr__(self) -> str:
        """
        Return a string representation of the block.
        """
        _str = f"\n\n#################\nlabel:\t{self.label}\nbbox:\t{self.bbox}\ncontent:\t{self.content}\n#################"
        return _str


def merge_formula_and_number(formula, formula_number):
    """
    Merge a formula and its formula number for display.

    Args:
        formula (str): The formula string.
        formula_number (str): The formula number string.

    Returns:
        str: The merged formula with tag.
    """
    formula = formula.replace("$$", "")
    merge_formula = r"{} \tag*{{{}}}".format(formula, formula_number)
    return f"$${merge_formula}$$"


def format_chart2table_func(block):
    lines_list = block.content.split("\n")
    # get header and rows
    header = lines_list[0].split("|")
    rows = [line.split("|") for line in lines_list[1:]]
    # construct html table
    html = "<table border=1 style='margin: auto; width: max-content;'>\n"
    html += (
        "  <thead><tr>"
        + "".join(
            f"<th style='text-align: center;'>{cell.strip()}</th>" for cell in header
        )
        + "</tr></thead>\n"
    )
    html += "  <tbody>\n"
    for row in rows:
        html += (
            "    <tr>"
            + "".join(
                f"<td style='text-align: center;'>{cell.strip()}</td>" for cell in row
            )
            + "</tr>\n"
        )
    html += "  </tbody>\n"
    html += "</table>"
    return html


def format_table_center_func(block):
    tabel_content = block.content

    tabel_content = tabel_content.replace(
        "<table>", "<table border=1 style='margin: auto; word-wrap: break-word;'>"
    )

    tabel_content = tabel_content.replace(
        "<th>", "<th style='text-align: center; word-wrap: break-word;'>"
    )
    tabel_content = tabel_content.replace(
        "<td>", "<td style='text-align: center; word-wrap: break-word;'>"
    )

    return tabel_content


def build_handle_funcs_dict(
    *,
    text_func,
    image_func,
    chart_func,
    table_func,
    formula_func,
    seal_func,
):
    """
    Build a dictionary mapping block labels to their formatting functions.

    Args:
        text_func: Function to format text blocks.
        image_func: Function to format image blocks.
        chart_func: Function to format chart blocks.
        table_func: Function to format table blocks.
        formula_func: Function to format formula blocks.
        seal_func: Function to format seal blocks.

    Returns:
        dict: A mapping from block label to handler function.
    """
    return {
        "paragraph_title": format_para_title_func,
        "abstract_title": format_title_func,
        "reference_title": format_title_func,
        "content_title": format_title_func,
        "doc_title": lambda block: f"# {block.content}".replace("-\n", "").replace(
            "\n", " "
        ),
        "table_title": text_func,
        "figure_title": text_func,
        "chart_title": text_func,
        "vision_footnote": lambda block: block.content.replace("\n\n", "\n").replace(
            "\n", "\n\n"
        ),
        "text": lambda block: block.content.replace("\n\n", "\n").replace("\n", "\n\n"),
        "ocr": lambda block: block.content.replace("\n\n", "\n").replace("\n", "\n\n"),
        "vertical_text": lambda block: block.content.replace("\n\n", "\n").replace(
            "\n", "\n\n"
        ),
        "reference_content": lambda block: block.content.replace("\n\n", "\n").replace(
            "\n", "\n\n"
        ),
        "abstract": partial(
            format_first_line_func,
            templates=["摘要", "abstract"],
            format_func=lambda l: f"## {l}\n",
            spliter=" ",
        ),
        "content": lambda block: block.content.replace("-\n", "  \n").replace(
            "\n", "  \n"
        ),
        "image": image_func,
        "chart": chart_func,
        "formula": formula_func,
        "display_formula": formula_func,
        "inline_formula": formula_func,
        "table": table_func,
        "reference": partial(
            format_first_line_func,
            templates=["参考文献", "references"],
            format_func=lambda l: f"## {l}",
            spliter="\n",
        ),
        "algorithm": lambda block: block.content.strip("\n"),
        "seal": seal_func,
        "spotting": lambda block: block.content,
        "number": format_text_plain_func,
        "footnote": format_text_plain_func,
        "header": format_text_plain_func,
        "header_image": image_func,
        "footer": format_text_plain_func,
        "footer_image": image_func,
        "aside_text": format_text_plain_func,
    }


@class_requires_deps("opencv-contrib-python")
class PaddleOCRVLResult(BaseCVResult, HtmlMixin, XlsxMixin, MarkdownMixin):
    """
    PaddleOCRVLResult class for holding and formatting OCR/VL parsing results.
    """

    def __init__(self, data) -> None:
        """
        Initializes a new instance of the class with the specified data.

        Args:
            data: The input data for the parsing result.
        """
        super().__init__(data)
        HtmlMixin.__init__(self)
        XlsxMixin.__init__(self)
        MarkdownMixin.__init__(self)
        JsonMixin.__init__(self)
        markdown_ignore_labels = self["model_settings"].get(
            "markdown_ignore_labels", []
        )
        self.skip_order_labels = [
            label for label in SKIP_ORDER_LABELS + markdown_ignore_labels
        ]

    def _to_img(self) -> dict[str, np.ndarray]:
        """
        Convert the parsing result to a dictionary of images.

        Returns:
            dict: Keys are names, values are numpy arrays (images).
        """

        res_img_dict = {}
        model_settings = self["model_settings"]
        if model_settings["use_doc_preprocessor"]:
            if isinstance(self["doc_preprocessor_res"], BaseResult):
                for key, value in self["doc_preprocessor_res"].img.items():
                    res_img_dict[key] = value
            if isinstance(self["doc_preprocessor_res"], list):
                for idx, doc_preprocessor_res in enumerate(
                    self["doc_preprocessor_res"]
                ):
                    if isinstance(doc_preprocessor_res, BaseResult):
                        for key, value in doc_preprocessor_res.img.items():
                            res_img_dict[f"{key}_{idx}"] = value
        if self["model_settings"]["use_layout_detection"]:
            if isinstance(self["layout_det_res"], BaseResult):
                res_img_dict["layout_det_res"] = self["layout_det_res"].img["res"]
            if isinstance(self["layout_det_res"], list):
                for idx, layout_res in enumerate(self["layout_det_res"]):
                    if isinstance(layout_res, BaseResult):
                        res_img_dict[f"layout_det_res_{idx}"] = layout_res.img["res"]

        if (
            self.get("spotting_res")
            and not isinstance(self["spotting_res"], list)
            and self.get("doc_preprocessor_res")
        ):
            boxes = self["spotting_res"]["rec_polys"]
            txts = self["spotting_res"]["rec_texts"]
            image = self["doc_preprocessor_res"]["output_img"][:, :, ::-1]
            h, w = image.shape[0:2]
            img_left = Image.fromarray(image)
            img_right = np.ones((h, w, 3), dtype=np.uint8) * 255
            random.seed(0)
            draw_left = ImageDraw.Draw(img_left)
            vis_font = SIMFANG_FONT
            for idx, (box, txt) in enumerate(zip(boxes, txts)):
                try:
                    color = (
                        random.randint(0, 255),
                        random.randint(0, 255),
                        random.randint(0, 255),
                    )
                    box = np.array(box)
                    if len(box) > 4:
                        pts = [(x, y) for x, y in box.tolist()]
                        draw_left.polygon(pts, outline=color, width=8, fill=color)
                        box = get_minarea_rect(box)
                        height = int(0.5 * (max(box[:, 1]) - min(box[:, 1])))
                        box[:2, 1] = np.mean(box[:, 1])
                        box[2:, 1] = np.mean(box[:, 1]) + min(20, height)
                    else:
                        box_pts = [(int(x), int(y)) for x, y in box.tolist()]
                        draw_left.polygon(box_pts, fill=color)
                    if isinstance(txt, tuple):
                        txt = txt[0]
                    img_right_text = draw_box_txt_fine((w, h), box, txt, vis_font.path)
                    pts = np.array(box, np.int32).reshape((-1, 1, 2))
                    cv2.polylines(img_right_text, [pts], True, color, 1)
                    img_right = cv2.bitwise_and(img_right, img_right_text)
                except:
                    continue

            img_left = Image.blend(Image.fromarray(image), img_left, 0.5)
            img_show = Image.new("RGB", (w * 2, h), (255, 255, 255))
            img_show.paste(img_left, (0, 0, w, h))
            img_show.paste(Image.fromarray(img_right), (w, 0, w * 2, h))

            res_img_dict["spotting_res_img"] = img_show

        return res_img_dict

    def _to_html(self) -> dict[str, str]:
        """
        Converts the prediction to its corresponding HTML representation.

        Returns:
            dict: The str type HTML representation result.
        """
        res_html_dict = {}
        if self.get("table_res_list") and len(self["table_res_list"]) > 0:
            for sno in range(len(self["table_res_list"])):
                table_res = self["table_res_list"][sno]
                table_region_id = table_res["table_region_id"]
                key = f"table_{table_region_id}"
                res_html_dict[key] = table_res.html["pred"]
        return res_html_dict

    def _to_xlsx(self) -> dict[str, str]:
        """
        Converts the prediction HTML to an XLSX file path.

        Returns:
            dict: The str type XLSX representation result.
        """
        res_xlsx_dict = {}
        if self.get("table_res_list") and len(self["table_res_list"]) > 0:
            for sno in range(len(self["table_res_list"])):
                table_res = self["table_res_list"][sno]
                table_region_id = table_res["table_region_id"]
                key = f"table_{table_region_id}"
                res_xlsx_dict[key] = table_res.xlsx["pred"]
        return res_xlsx_dict

    def _to_str(self, *args, **kwargs) -> dict[str, str]:
        """
        Converts the instance's attributes to a dictionary and then to a string.

        Args:
            *args: Additional positional arguments passed to the base class method.
            **kwargs: Additional keyword arguments passed to the base class method.

        Returns:
            dict: A dictionary with the instance's attributes converted to strings.
        """
        data = {}
        data["input_path"] = self["input_path"]
        data["page_index"] = self["page_index"]
        data["page_count"] = self["page_count"]
        data["width"] = self["width"]
        data["height"] = self["height"]
        model_settings = self["model_settings"]
        data["model_settings"] = model_settings
        if self["model_settings"]["use_doc_preprocessor"]:
            if isinstance(self["doc_preprocessor_res"], BaseResult):
                data["doc_preprocessor_res"] = self["doc_preprocessor_res"].str["res"]
            else:
                data["doc_preprocessor_res"] = self["doc_preprocessor_res"]
        if self["model_settings"]["use_layout_detection"]:
            if isinstance(self["layout_det_res"], BaseResult):
                data["layout_det_res"] = self["layout_det_res"].str["res"]
            else:
                data["layout_det_res"] = self["layout_det_res"]
        parsing_res_list = self["parsing_res_list"]
        parsing_res_list = [
            {
                "block_label": parsing_res.label,
                "block_content": parsing_res.content,
                "block_bbox": parsing_res.bbox,
            }
            for parsing_res in parsing_res_list
        ]
        data["parsing_res_list"] = parsing_res_list
        return JsonMixin._to_str(data, *args, **kwargs)

    def _to_json(self, *args, **kwargs) -> dict[str, str]:
        """
        Converts the object's data to a JSON dictionary.

        Args:
            *args: Positional arguments passed to the JsonMixin._to_json method.
            **kwargs: Keyword arguments passed to the JsonMixin._to_json method.

        Returns:
            dict: A dictionary containing the object's data in JSON format.
        """
        _keep_img = kwargs.pop("keep_img", False)

        data = {}
        data["input_path"] = self["input_path"]
        data["page_index"] = self["page_index"]
        data["page_count"] = self["page_count"]
        data["width"] = self["width"]
        data["height"] = self["height"]
        model_settings = self["model_settings"]
        data["model_settings"] = model_settings
        use_seal_recognition = self["model_settings"].get("use_seal_recognition", False)
        if self["model_settings"].get("format_block_content", False):
            original_image_width = self["width"]
            use_ocr_for_image_block = self["model_settings"].get(
                "use_ocr_for_image_block", False
            )
            format_text_func = lambda block: format_centered_by_html(
                format_text_plain_func(block)
            )
            format_image_func = lambda block: format_centered_by_html(
                format_image_scaled_by_html_func(
                    block,
                    original_image_width=original_image_width,
                    show_ocr_content=use_ocr_for_image_block,
                ),
                remove_symbol=not use_ocr_for_image_block,
            )

            format_seal_func = lambda block: format_centered_by_html(
                format_image_scaled_by_html_func(
                    block,
                    original_image_width=original_image_width,
                    show_ocr_content=True,
                ),
                remove_symbol=use_seal_recognition,
            )

            if self["model_settings"].get("use_chart_recognition", False):
                format_chart_func = format_chart2table_func
            else:
                format_chart_func = format_image_func

            if not self["model_settings"].get("use_layout_detection", False):
                format_seal_func = format_text_func

            format_table_func = lambda block: "\n" + format_table_center_func(block)
            format_formula_func = lambda block: block.content

            handle_funcs_dict = build_handle_funcs_dict(
                text_func=format_text_func,
                image_func=format_image_func,
                chart_func=format_chart_func,
                table_func=format_table_func,
                formula_func=format_formula_func,
                seal_func=format_seal_func,
            )

        parsing_res_list = self["parsing_res_list"]
        parsing_res_list_json = []
        order_index = 1
        for idx, parsing_res in enumerate(parsing_res_list):
            label = parsing_res.label
            if label not in self.skip_order_labels:
                order = order_index
                order_index += 1
            else:
                order = None
            res_dict = {
                "block_label": parsing_res.label,
                "block_content": parsing_res.content,
                "block_bbox": parsing_res.bbox,
                "block_id": idx,
                "block_order": order,
                "group_id": (
                    parsing_res.group_id if parsing_res.group_id is not None else idx
                ),
            }
            if (
                hasattr(parsing_res, "global_block_id")
                and parsing_res.global_block_id is not None
            ):
                res_dict["global_block_id"] = parsing_res.global_block_id
            if (
                hasattr(parsing_res, "global_group_id")
                and parsing_res.global_group_id is not None
            ):
                res_dict["global_group_id"] = parsing_res.global_group_id
            if parsing_res.polygon_points is not None:
                res_dict["block_polygon_points"] = parsing_res.polygon_points

            if _keep_img and parsing_res.image is not None:
                res_dict["image"] = parsing_res.image

            if self["model_settings"].get("format_block_content", False):
                if handle_funcs_dict.get(parsing_res.label):
                    res_dict["block_content"] = handle_funcs_dict[parsing_res.label](
                        parsing_res
                    )
                else:
                    res_dict["block_content"] = parsing_res.content

            parsing_res_list_json.append(res_dict)
        data["parsing_res_list"] = parsing_res_list_json
        if self.get("spotting_res"):
            if isinstance(self["spotting_res"], list):
                data["spotting_res"] = [res for res in self["spotting_res"]]
            else:
                data["spotting_res"] = self["spotting_res"]
        if self["model_settings"]["use_doc_preprocessor"]:
            if isinstance(self["doc_preprocessor_res"], BaseResult):
                data["doc_preprocessor_res"] = self["doc_preprocessor_res"].json["res"]
            elif isinstance(self["doc_preprocessor_res"], list):
                doc_preprocessor_res = []
                for res in self["doc_preprocessor_res"]:
                    if isinstance(res, BaseResult):
                        doc_preprocessor_res.append(res.json["res"])
                    else:
                        doc_preprocessor_res.append(res)
                data["doc_preprocessor_res"] = doc_preprocessor_res
            else:
                data["doc_preprocessor_res"] = self["doc_preprocessor_res"]
        if self["model_settings"]["use_layout_detection"]:
            if isinstance(self["layout_det_res"], BaseResult):
                data["layout_det_res"] = self["layout_det_res"].json["res"]
            elif isinstance(self["layout_det_res"], list):
                layout_det_res = []
                for res in self["layout_det_res"]:
                    if isinstance(res, BaseResult):
                        layout_det_res.append(res.json["res"])
                    else:
                        layout_det_res.append(res)
                data["layout_det_res"] = layout_det_res
            else:
                data["layout_det_res"] = self["layout_det_res"]

        return JsonMixin._to_json(data, *args, **kwargs)

    def _to_markdown(self, pretty=True, show_formula_number=False) -> dict:
        """
        Save the parsing result to a Markdown file.

        Args:
            pretty (Optional[bool]): whether to pretty markdown by HTML, default by True.
            show_formula_number (bool): whether to show formula numbers.

        Returns:
            dict: Markdown information with text and images.
        """

        use_ocr_for_image_block = self["model_settings"].get(
            "use_ocr_for_image_block", False
        )
        use_seal_recognition = self["model_settings"].get("use_seal_recognition", False)
        if isinstance(self["width"], list):
            original_image_width = self["width"][0]
        else:
            original_image_width = self["width"]

        if pretty:
            format_text_func = lambda block: format_centered_by_html(
                format_text_plain_func(block)
            )
            format_image_func = lambda block: format_centered_by_html(
                format_image_scaled_by_html_func(
                    block,
                    original_image_width=original_image_width,
                    show_ocr_content=use_ocr_for_image_block,
                ),
                remove_symbol=not use_ocr_for_image_block,
            )
            format_seal_func = lambda block: format_centered_by_html(
                format_image_scaled_by_html_func(
                    block,
                    original_image_width=original_image_width,
                    show_ocr_content=use_seal_recognition,
                ),
                remove_symbol=False,
            )
        else:
            format_text_func = lambda block: block.content
            format_image_func = lambda block: format_image_plain_func(
                block, use_ocr_for_image_block
            )
            format_seal_func = lambda block: format_image_plain_func(
                block, use_seal_recognition
            )

        format_chart_func = (
            format_chart2table_func
            if self["model_settings"]["use_chart_recognition"]
            else format_image_func
        )

        if not self["model_settings"].get("use_layout_detection", False):
            format_seal_func = format_text_func

        if pretty:
            format_table_func = lambda block: "\n" + format_table_center_func(block)
        else:
            format_table_func = lambda block: simplify_table_func("\n" + block.content)

        format_formula_func = lambda block: block.content

        handle_funcs_dict = build_handle_funcs_dict(
            text_func=format_text_func,
            image_func=format_image_func,
            chart_func=format_chart_func,
            table_func=format_table_func,
            formula_func=format_formula_func,
            seal_func=format_seal_func,
        )
        for label in self["model_settings"].get("markdown_ignore_labels", []):
            handle_funcs_dict.pop(label, None)

        markdown_content = ""
        markdown_info = {}
        markdown_info["markdown_images"] = {}
        for idx, block in enumerate(self["parsing_res_list"]):
            label = block.label
            if block.image is not None:
                markdown_info["markdown_images"][block.image["path"]] = block.image[
                    "img"
                ]
            handle_func = handle_funcs_dict.get(label, None)
            if (
                show_formula_number
                and (label == "display_formula" or label == "formula")
                and idx != len(self["parsing_res_list"]) - 1
            ):
                next_block = self["parsing_res_list"][idx + 1]
                next_block_label = next_block.label
                if next_block_label == "formula_number":
                    block.content = merge_formula_and_number(
                        block.content, next_block.content
                    )
            if handle_func:
                markdown_content += (
                    "\n\n" + handle_func(block)
                    if markdown_content
                    else handle_func(block)
                )

        markdown_info["page_index"] = self["page_index"]
        markdown_info["input_path"] = self["input_path"]
        markdown_info["markdown_texts"] = markdown_content
        for img in self["imgs_in_doc"]:
            markdown_info["markdown_images"][img["path"]] = img["img"]

        return markdown_info


class PaddleOCRVLPagesResult(PaddleOCRVLResult):
    def save_to_img(self, *args, **kwargs):
        logging.warning(
            f"The result of multi-pages don't support to save as image format!"
        )
        return None

    def save_to_html(self, *args, **kwargs):
        logging.warning(
            f"The result of multi-pages don't support to save as html format!"
        )
        return None

    def save_to_xlsx(self, *args, **kwargs):
        logging.warning(
            f"The result of multi-pages don't support to save as xlsx format!"
        )
        return None

    def _to_markdown(self, pretty=True, show_formula_number=False) -> dict:
        """
        Save the parsing result to a Markdown file.

        Args:
            pretty (Optional[bool]): whether to pretty markdown by HTML, default by True.
            show_formula_number (bool): whether to show formula numbers.

        Returns:
            dict: Markdown information with text and images.
        """

        use_ocr_for_image_block = self["model_settings"].get(
            "use_ocr_for_image_block", False
        )
        use_seal_recognition = self["model_settings"].get("use_seal_recognition", False)
        if isinstance(self["width"], list):
            original_image_width = self["width"][0]
        else:
            original_image_width = self["width"]

        if pretty:
            format_text_func = lambda block: format_centered_by_html(
                format_text_plain_func(block)
            )
            format_image_func = lambda block: format_centered_by_html(
                format_image_scaled_by_html_func(
                    block,
                    original_image_width=original_image_width,
                    show_ocr_content=use_ocr_for_image_block,
                ),
                remove_symbol=not use_ocr_for_image_block,
            )
            format_seal_func = lambda block: format_centered_by_html(
                format_image_scaled_by_html_func(
                    block,
                    original_image_width=original_image_width,
                    show_ocr_content=use_seal_recognition,
                ),
                remove_symbol=False,
            )
        else:
            format_text_func = lambda block: block.content
            format_image_func = lambda block: format_image_plain_func(
                block, use_ocr_for_image_block
            )
            format_seal_func = lambda block: format_image_plain_func(
                block, use_seal_recognition
            )

        format_chart_func = (
            format_chart2table_func
            if self["model_settings"]["use_chart_recognition"]
            else format_image_func
        )

        if pretty:
            format_table_func = lambda block: "\n" + format_table_center_func(block)
        else:
            format_table_func = lambda block: simplify_table_func("\n" + block.content)

        format_formula_func = lambda block: block.content

        handle_funcs_dict = build_handle_funcs_dict(
            text_func=format_text_func,
            image_func=format_image_func,
            chart_func=format_chart_func,
            table_func=format_table_func,
            formula_func=format_formula_func,
            seal_func=format_seal_func,
        )
        for label in self["model_settings"].get("markdown_ignore_labels", []):
            handle_funcs_dict.pop(label, None)

        markdown_content = ""
        markdown_info = {}
        markdown_info["markdown_images"] = {}
        for idx, block in enumerate(self["parsing_res_list"]):
            label = block.label
            if block.image is not None:
                markdown_info["markdown_images"][block.image["path"]] = block.image[
                    "img"
                ]
            handle_func = handle_funcs_dict.get(label, None)
            if (
                show_formula_number
                and (label == "display_formula" or label == "formula")
                and idx != len(self["parsing_res_list"]) - 1
            ):
                next_block = self["parsing_res_list"][idx + 1]
                next_block_label = next_block.label
                if next_block_label == "formula_number":
                    block.content = merge_formula_and_number(
                        block.content, next_block.content
                    )
            if handle_func:
                markdown_content += (
                    "\n\n" + handle_func(block)
                    if markdown_content
                    else handle_func(block)
                )

        markdown_info["page_index"] = self["page_index"]
        markdown_info["input_path"] = self["input_path"]
        markdown_info["markdown_texts"] = markdown_content
        for img in self["imgs_in_doc"]:
            markdown_info["markdown_images"][img["path"]] = img["img"]

        return markdown_info



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List

from .....utils.deps import function_requires_deps, is_dep_available
from ...infra import utils as serving_utils
from ...infra.config import AppConfig
from ...infra.models import AIStudioResultResponse
from ...schemas.paddleocr_vl import (
    INFER_ENDPOINT,
    RESTRUCTURE_PAGES_ENDPOINT,
    InferRequest,
    InferResult,
    RestructurePagesRequest,
    RestructurePagesResult,
)
from .._app import create_app, primary_operation
from ._common import common
from ._common import ocr as ocr_common

if is_dep_available("fastapi"):
    from fastapi import FastAPI


@function_requires_deps("fastapi")
def create_pipeline_app(pipeline: Any, app_config: AppConfig) -> "FastAPI":
    app, ctx = create_app(
        pipeline=pipeline, app_config=app_config, app_aiohttp_session=True
    )

    ocr_common.update_app_context(ctx)

    @primary_operation(
        app,
        INFER_ENDPOINT,
        "infer",
    )
    async def _infer(
        request: InferRequest,
    ) -> AIStudioResultResponse[InferResult]:
        pipeline = ctx.pipeline

        log_id = request.logId if request.logId else serving_utils.generate_log_id()
        visualize_enabled = (
            request.visualize if request.visualize is not None else ctx.config.visualize
        )
        images, data_info = await ocr_common.get_images(request, ctx)

        result = await pipeline.infer(
            images,
            use_doc_orientation_classify=request.useDocOrientationClassify,
            use_doc_unwarping=request.useDocUnwarping,
            use_layout_detection=request.useLayoutDetection,
            use_chart_recognition=request.useChartRecognition,
            use_seal_recognition=request.useSealRecognition,
            use_ocr_for_image_block=request.useOcrForImageBlock,
            layout_threshold=request.layoutThreshold,
            layout_nms=request.layoutNms,
            layout_unclip_ratio=request.layoutUnclipRatio,
            layout_merge_bboxes_mode=request.layoutMergeBboxesMode,
            layout_shape_mode=request.layoutShapeMode,
            prompt_label=request.promptLabel,
            format_block_content=request.formatBlockContent,
            repetition_penalty=request.repetitionPenalty,
            temperature=request.temperature,
            top_p=request.topP,
            min_pixels=request.minPixels,
            max_pixels=request.maxPixels,
            max_new_tokens=request.maxNewTokens,
            merge_layout_blocks=request.mergeLayoutBlocks,
            markdown_ignore_labels=request.markdownIgnoreLabels,
            vlm_extra_args=request.vlmExtraArgs,
        )

        orig_result = result
        if request.restructurePages:
            result = await serving_utils.call_async(
                pipeline.pipeline.restructure_pages,
                result,
                merge_tables=request.mergeTables,
                relevel_titles=request.relevelTitles,
                concatenate_pages=False,
            )
            result = list(result)

        layout_parsing_results: List[Dict[str, Any]] = []
        for i, (img, item) in enumerate(zip(images, result)):
            pruned_res = common.prune_result(item.json["res"])
            # XXX
            md_data = item._to_markdown(
                pretty=request.prettifyMarkdown,
                show_formula_number=request.showFormulaNumber,
            )
            md_text = md_data["markdown_texts"]
            md_imgs = await serving_utils.call_async(
                common.postprocess_images,
                md_data["markdown_images"],
                log_id,
                filename_template=f"markdown_{i}/{{key}}",
                file_storage=ctx.extra["file_storage"],
                return_urls=ctx.extra["return_img_urls"],
                max_img_size=ctx.extra["max_output_img_size"],
            )
            if visualize_enabled:
                imgs = {
                    "input_img": img,
                    **item.img,
                }
                imgs = await serving_utils.call_async(
                    common.postprocess_images,
                    imgs,
                    log_id,
                    filename_template=f"{{key}}_{i}.jpg",
                    file_storage=ctx.extra["file_storage"],
                    return_urls=ctx.extra["return_img_urls"],
                    max_img_size=ctx.extra["max_output_img_size"],
                )
            else:
                imgs = {}
            layout_parsing_results.append(
                dict(
                    prunedResult=pruned_res,
                    markdown=dict(
                        text=md_text,
                        images=md_imgs,
                    ),
                    outputImages=(
                        {k: v for k, v in imgs.items() if k != "input_img"}
                        if imgs
                        else None
                    ),
                    inputImage=imgs.get("input_img"),
                )
            )

        return AIStudioResultResponse[InferResult](
            logId=log_id,
            result=InferResult(
                layoutParsingResults=layout_parsing_results,
                dataInfo=data_info,
            ),
        )

    @primary_operation(
        app,
        RESTRUCTURE_PAGES_ENDPOINT,
        "restructurePages",
    )
    async def _restructure_pages(
        request: RestructurePagesRequest,
    ) -> AIStudioResultResponse[RestructurePagesResult]:
        def _to_original_result(pruned_res, page_index):
            res = {**pruned_res, "input_path": "", "page_index": page_index}
            orig_res = {"res": res}
            return orig_res

        pipeline = ctx.pipeline

        log_id = request.logId if request.logId else serving_utils.generate_log_id()

        original_results = []
        markdown_images = {}
        for i, page in enumerate(request.pages):
            orig_res = _to_original_result(page.prunedResult, i)
            original_results.append(orig_res)
            if request.concatenatePages:
                markdown_images.update(page.markdownImages)

        restructured_results = await serving_utils.call_async(
            pipeline.pipeline.restructure_pages,
            original_results,
            merge_tables=request.mergeTables,
            relevel_titles=request.relevelTitles,
            concatenate_pages=request.concatenatePages,
        )
        restructured_results = list(restructured_results)

        layout_parsing_results = []
        if request.concatenatePages:
            layout_parsing_result = {}
            layout_parsing_result["prunedResult"] = common.prune_result(
                restructured_results[0].json["res"]
            )
            # XXX
            md_data = restructured_results[0]._to_markdown(
                pretty=request.prettifyMarkdown,
                show_formula_number=request.showFormulaNumber,
            )
            layout_parsing_result["markdown"] = dict(
                text=md_data["markdown_texts"],
                images=markdown_images,
            )
            layout_parsing_results.append(layout_parsing_result)
        else:
            for new_res, old_page in zip(restructured_results, request.pages):
                layout_parsing_result = {}
                layout_parsing_result["prunedResult"] = common.prune_result(
                    new_res.json["res"]
                )
                # XXX
                md_data = new_res._to_markdown(
                    pretty=request.prettifyMarkdown,
                    show_formula_number=request.showFormulaNumber,
                )
                layout_parsing_result["markdown"] = dict(
                    text=md_data["markdown_texts"],
                    images=old_page.markdownImages,
                )
                layout_parsing_results.append(layout_parsing_result)

        return AIStudioResultResponse[RestructurePagesResult](
            logId=log_id,
            result=RestructurePagesResult(
                layoutParsingResults=layout_parsing_results,
            ),
        )

    return app



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ....modules.text_to_speech_vocoder.model_list import MODELS
from ...common.batch_sampler import AudioBatchSampler
from ..base import BasePredictor
from .result import PwganResult


class PwganPredictor(BasePredictor):

    entities = MODELS

    def __init__(self, *args, **kwargs):
        """Initializes FastspeechPredictor.

        Args:
            *args: Arbitrary positional arguments passed to the superclass.
            **kwargs: Arbitrary keyword arguments passed to the superclass.
        """
        super().__init__(*args, **kwargs)
        self.infer = self.create_static_infer()

    def _build_batch_sampler(self):
        """Builds and returns an AudioBatchSampler instance.

        Returns:
            AudioBatchSampler: An instance of AudioBatchSampler.
        """
        return AudioBatchSampler()

    def _get_result_class(self):
        """Returns the result class, PwganResult.

        Returns:
            type: The PwganResult class.
        """
        return PwganResult

    def process(self, batch_data):
        """
        Process a batch of data through the preprocessing, inference, and postprocessing.

        Args:
            batch_data (List[Union[str], ...]): A batch of input phone data.

        Returns:
            dict: A dictionary containing the input path and result. The result include the output pinyin dict.
        """
        input_data = batch_data[0]
        if type(input_data) is str:
            mel = np.load(input_data)
        else:
            mel = input_data
        wav = self.infer([mel])
        result = np.array(wav).reshape(1, -1)
        return {
            "result": result,
        }



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle

from ....utils.benchmark import add_inference_operations, benchmark
from ...common.transformers.transformers import PretrainedConfig, PretrainedModel
from .slanext_modules.rec_vary_vit import Vary_VIT_B
from .slanext_modules.table_att_head import SLAHead

__all__ = ["SLANeXt"]


class SLANeXtConfig(PretrainedConfig):
    def __init__(
        self,
        backbone,
        SLAHead,
    ):
        if backbone["name"] == "Vary_VIT_B":
            self.image_size = backbone["image_size"]
            self.encoder_embed_dim = backbone["encoder_embed_dim"]
            self.encoder_depth = backbone["encoder_depth"]
            self.encoder_num_heads = backbone["encoder_num_heads"]
            self.encoder_global_attn_indexes = backbone["encoder_global_attn_indexes"]
        else:
            raise RuntimeError(
                f"There is no dynamic graph implementation for backbone {backbone['name']}."
            )
        self.out_channels = SLAHead["out_channels"]
        self.hidden_size = SLAHead["hidden_size"]
        self.max_text_length = SLAHead["max_text_length"]
        self.loc_reg_num = SLAHead["loc_reg_num"]
        self.tensor_parallel_degree = 1


class SLANeXt(PretrainedModel):

    config_class = SLANeXtConfig

    def __init__(self, config: SLANeXtConfig):
        super().__init__(config)
        self.backbone = Vary_VIT_B(
            image_size=self.config.image_size,
            encoder_embed_dim=self.config.encoder_embed_dim,
            encoder_depth=self.config.encoder_depth,
            encoder_num_heads=self.config.encoder_num_heads,
            encoder_global_attn_indexes=self.config.encoder_global_attn_indexes,
        )
        self.head = SLAHead(
            in_channels=self.backbone.out_channels,
            out_channels=self.config.out_channels,
            hidden_size=self.config.hidden_size,
            max_text_length=self.config.max_text_length,
            loc_reg_num=self.config.loc_reg_num,
        )

    add_inference_operations("slanext_forward")

    @benchmark.timeit_with_options(name="slanext_forward")
    def forward(self, x):
        x = paddle.to_tensor(x[0])
        x = self.backbone(x)
        x = self.head(x)
        return [x["loc_preds"], x["structure_probs"]]

    def get_transpose_weight_keys(self):
        transpose_keys = ["mlp.lin2", "attn.qkv", "mlp.lin1"]
        need_to_transpose = []
        all_weight_keys = []
        for name, param in self.backbone.named_parameters():
            all_weight_keys.append("backbone." + name)
        for i in range(len(all_weight_keys)):
            for j in range(len(transpose_keys)):
                if (transpose_keys[j] in all_weight_keys[i]) and (
                    "bias" not in all_weight_keys[i]
                ):
                    need_to_transpose.append(all_weight_keys[i])
        return need_to_transpose

    def get_hf_state_dict(self, *args, **kwargs):

        model_state_dict = self.state_dict(*args, **kwargs)

        hf_state_dict = {}
        for old_key, value in model_state_dict.items():
            if "_mean" in old_key:
                new_key = old_key.replace("_mean", "running_mean")
            elif "_variance" in old_key:
                new_key = old_key.replace("_variance", "running_var")
            else:
                new_key = old_key
            hf_state_dict[new_key] = value

        return hf_state_dict

    def set_hf_state_dict(self, state_dict, *args, **kwargs):

        key_mapping = {}
        for old_key in list(state_dict.keys()):
            if "running_mean" in old_key:
                key_mapping[old_key] = old_key.replace("running_mean", "_mean")
            elif "running_var" in old_key:
                key_mapping[old_key] = old_key.replace("running_var", "_variance")

        for old_key, new_key in key_mapping.items():
            state_dict[new_key] = state_dict.pop(old_key)

        return self.set_state_dict(state_dict, *args, **kwargs)



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml


def load_backend_config(config_path):
    with open(config_path, "r", encoding="utf-8") as f:
        config = yaml.safe_load(f)
    return config


def update_backend_config(config, overrides):
    for k, v in overrides.items():
        config[k] = v


def set_config_defaults(config, defaults):
    for k, v in defaults.items():
        if k not in config:
            config[k] = v


def backend_config_to_args(config, convert_underscores_to_dashes=True):
    # Limited support
    args = []
    for k, v in config.items():
        if convert_underscores_to_dashes:
            k = k.replace("_", "-")
        opt = "--" + k
        args.append(opt)
        if not isinstance(v, bool):
            args.append(str(v))
    return args



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ....utils import logging
from .base_batch_sampler import BaseBatchSampler


class TextBatchSampler(BaseBatchSampler):
    def __init__(self):
        """Initializes the BaseBatchSampler.

        Args:
            batch_size (int, optional): The size of each batch. Only support 1.
        """
        super().__init__()
        self.batch_size = 1

    def sample(self, inputs):
        """Generate list of input file path.

        Args:
            inputs (str): file path.

        Yields:
            list: list of file path.
        """
        if isinstance(inputs, str):
            yield [inputs]
        else:
            logging.warning(
                f"Not supported input data type! Only `str` are supported, but got: {input}."
            )

    @BaseBatchSampler.batch_size.setter
    def batch_size(self, batch_size):
        """Sets the batch size.

        Args:
            batch_size (int): The batch size to set.

        Raises:
            Warning: If the batch size is not equal 1.
        """
        # only support batch size 1
        if batch_size != 1:
            logging.warning(
                f"audio batch sampler only support batch size 1, but got {batch_size}."
            )
        else:
            self._batch_size = batch_size



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...common.transformers.transformers import PretrainedConfig

DEFAULT_CONFIG = {
    "model_name": "UVDoc",
    "num_filter": 32,
    "in_channels": 3,
    "kernel_size": 5,
    "stride": [1, 2, 2, 2],
    "map_num": [1, 2, 4, 8, 16],
    "block_nums": [3, 4, 6, 3],
    "dilation_values": {
        "bridge_1": 1,
        "bridge_2": 2,
        "bridge_3": 5,
        "bridge_4": [8, 3, 2],
        "bridge_5": [12, 7, 4],
        "bridge_6": [18, 12, 6],
    },
    "padding_mode": "reflect",
    "upsample_size": [712, 488],
    "upsample_mode": "bilinear",
}


class UVDocNetConfig(PretrainedConfig):
    model_type = "uvdoc"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.model_name = kwargs.get("model_name", DEFAULT_CONFIG["model_name"])
        self.num_filter = kwargs.get("num_filter", DEFAULT_CONFIG["num_filter"])
        self.in_channels = kwargs.get("in_channels", DEFAULT_CONFIG["in_channels"])
        self.kernel_size = kwargs.get("kernel_size", DEFAULT_CONFIG["kernel_size"])
        self.stride = kwargs.get("stride", DEFAULT_CONFIG["stride"])
        self.map_num = kwargs.get("map_num", DEFAULT_CONFIG["map_num"])
        self.block_nums = kwargs.get("block_nums", DEFAULT_CONFIG["block_nums"])
        self.dilation_values = kwargs.get(
            "dilation_values", DEFAULT_CONFIG["dilation_values"]
        )
        self.padding_mode = kwargs.get("padding_mode", DEFAULT_CONFIG["padding_mode"])
        self.upsample_size = kwargs.get(
            "upsample_size", DEFAULT_CONFIG["upsample_size"]
        )
        self.upsample_mode = kwargs.get(
            "upsample_mode", DEFAULT_CONFIG["upsample_mode"]
        )



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import logging
from .... import UltraInferModel, ModelFormat
from .... import c_lib_wrap as C
from ...common import ProcessorManager
from ...detection.ppdet import PicoDet


class PPShiTuV2Detector(PicoDet):
    """Detect main body from an input image."""

    ...


class PPShiTuV2RecognizerPreprocessor(ProcessorManager):
    def __init__(self, config_file):
        """Create a preprocessor for PPShiTuV2Recognizer from configuration file

        :param config_file: (str)Path of configuration file, e.g PPLCNet/inference_cls.yaml
        """
        super(PPShiTuV2RecognizerPreprocessor, self).__init__()
        self._manager = C.vision.classification.PPShiTuV2RecognizerPreprocessor(
            config_file
        )

    def disable_normalize(self):
        """
        This function will disable normalize in preprocessing step.
        """
        self._manager.disable_normalize()

    def disable_permute(self):
        """
        This function will disable hwc2chw in preprocessing step.
        """
        self._manager.disable_permute()

    def initial_resize_on_cpu(self, v):
        """
        When the initial operator is Resize, and input image size is large,
        maybe it's better to run resize on CPU, because the HostToDevice memcpy
        is time consuming. Set this True to run the initial resize on CPU.
        :param: v: True or False
        """
        self._manager.initial_resize_on_cpu(v)


class PPShiTuV2RecognizerPostprocessor:
    def __init__(self, topk=1):
        """Create a postprocessor for PPShiTuV2Recognizer"""
        self._postprocessor = C.vision.classification.PPShiTuV2RecognizerPostprocessor()

    def run(self, runtime_results):
        """Postprocess the runtime results for PPShiTuV2Recognizer

        :param: runtime_results: (list of FDTensor)The output FDTensor results from runtime
        :return: list of ClassifyResult, the feature vector is ClassifyResult.feature (If the runtime_results is predict by batched samples, the length of this list equals to the batch size)
        """
        return self._postprocessor.run(runtime_results)


class PPShiTuV2Recognizer(UltraInferModel):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a image PPShiTuV2Recognizer model exported by PaddleClas.

        :param model_file: (str)Path of model file, e.g PPLCNet/inference.pdmodel
        :param params_file: (str)Path of parameters file, e.g PPLCNet/inference.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str) Path of configuration file for deploy, e.g PPLCNet/inference_cls.yaml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPShiTuV2Recognizer, self).__init__(runtime_option)
        self._model = C.vision.classification.PPShiTuV2Recognizer(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PPShiTuV2Recognizer model initialize failed."

    def clone(self):
        """Clone PPShiTuV2Recognizer object

        :return: a new PPShiTuV2Recognizer object
        """

        class PPShiTuV2RecognizerCloneModel(PPShiTuV2Recognizer):
            def __init__(self, model):
                self._model = model

        clone_model = PPShiTuV2RecognizerCloneModel(self._model.clone())
        return clone_model

    def predict(self, im):
        """Extract feature from an input image

        :param im: (numpy.ndarray) The input image data, a 3-D array with layout HWC, BGR format
        :return: ClassifyResult
        """

        return self._model.predict(im)

    def batch_predict(self, images):
        """Extract features from a batch of input image

        :param im: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return list of ClassifyResult, the feature vector is ClassifyResult.feature
        """

        return self._model.batch_predict(images)

    @property
    def preprocessor(self):
        """Get PPShiTuV2RecognizerPreprocessor object of the loaded model

        :return PPShiTuV2RecognizerPreprocessor
        """
        return self._model.preprocessor

    @property
    def postprocessor(self):
        """Get PPShiTuV2RecognizerPostprocessor object of the loaded model

        :return PPShiTuV2RecognizerPostprocessor
        """
        return self._model.postprocessor



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tqdm import trange
import numpy as np
import collections
import os
import math
import time


def eval_segmentation(model, data_dir, batch_size=1):
    import cv2
    from .utils import Cityscapes
    from .utils import f1_score, calculate_area, mean_iou, accuracy, kappa

    assert os.path.isdir(data_dir), "The image_file_path:{} is not a directory.".format(
        data_dir
    )
    eval_dataset = Cityscapes(dataset_root=data_dir, mode="val")
    file_list = eval_dataset.file_list
    image_num = eval_dataset.num_samples
    num_classes = eval_dataset.num_classes
    intersect_area_all = 0
    pred_area_all = 0
    label_area_all = 0
    conf_mat_all = []
    twenty_percent_image_num = math.ceil(image_num * 0.2)
    start_time = 0
    end_time = 0
    average_inference_time = 0
    im_list = []
    label_list = []
    for image_label_path, i in zip(
        file_list, trange(image_num, desc="Inference Progress")
    ):
        if i == twenty_percent_image_num:
            start_time = time.time()
        im = cv2.imread(image_label_path[0])
        label = cv2.imread(image_label_path[1], cv2.IMREAD_GRAYSCALE)
        label_list.append(label)
        if batch_size == 1:
            result = model.predict(im)
            results = [result]
        else:
            im_list.append(im)
            # If the batch_size is not satisfied, the remaining pictures are formed into a batch
            if (i + 1) % batch_size != 0 and i != image_num - 1:
                continue
            results = model.batch_predict(im_list)
        if i == image_num - 1:
            end_time = time.time()
            average_inference_time = round(
                (end_time - start_time) / (image_num - twenty_percent_image_num), 4
            )
        for result, label in zip(results, label_list):
            pred = np.array(result.label_map).reshape(result.shape[0], result.shape[1])
            intersect_area, pred_area, label_area = calculate_area(
                pred, label, num_classes
            )
            intersect_area_all = intersect_area_all + intersect_area
            pred_area_all = pred_area_all + pred_area
            label_area_all = label_area_all + label_area
        im_list.clear()
        label_list.clear()

    class_iou, miou = mean_iou(intersect_area_all, pred_area_all, label_area_all)
    class_acc, oacc = accuracy(intersect_area_all, pred_area_all)
    kappa_res = kappa(intersect_area_all, pred_area_all, label_area_all)
    category_f1score = f1_score(intersect_area_all, pred_area_all, label_area_all)

    eval_metrics = collections.OrderedDict(
        zip(
            [
                "miou",
                "category_iou",
                "oacc",
                "category_acc",
                "kappa",
                "category_F1-score",
                "average_inference_time(s)",
            ],
            [
                miou,
                class_iou,
                oacc,
                class_acc,
                kappa_res,
                category_f1score,
                average_inference_time,
            ],
        )
    )
    return eval_metrics



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...common.transformers.transformers import PretrainedConfig

DEFAULT_CONFIG = {
    "model_name": "PP-LCNet_x1_0_doc_ori",
    "scale": 1.0,
    "class_num": 4,
    "stride_list": [2, 2, 2, 2, 2],
    "dropout_prob": 0.2,
    "class_expand": 1280,
    "use_last_conv": True,
    "act": "hardswish",
    "reduction": 4,
    "lr_mult_list": [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
    "net_config": {
        "blocks2": [[3, 16, 32, 1, False]],
        "blocks3": [[3, 32, 64, 2, False], [3, 64, 64, 1, False]],
        "blocks4": [[3, 64, 128, 2, False], [3, 128, 128, 1, False]],
        "blocks5": [
            [3, 128, 256, 2, False],
            [5, 256, 256, 1, False],
            [5, 256, 256, 1, False],
            [5, 256, 256, 1, False],
            [5, 256, 256, 1, False],
            [5, 256, 256, 1, False],
        ],
        "blocks6": [[5, 256, 512, 2, True], [5, 512, 512, 1, True]],
    },
}


class PPLCNetConfig(PretrainedConfig):
    model_type = "cls"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.model_name = kwargs.get("model_name", DEFAULT_CONFIG["model_name"])
        self.scale = kwargs.get("scale", DEFAULT_CONFIG["scale"])
        self.class_num = kwargs.get("class_num", DEFAULT_CONFIG["class_num"])
        self.stride_list = kwargs.get("stride_list", DEFAULT_CONFIG["stride_list"])
        self.reduction = kwargs.get("reduction", DEFAULT_CONFIG["reduction"])
        self.dropout_prob = kwargs.get("dropout_prob", DEFAULT_CONFIG["dropout_prob"])
        self.class_expand = kwargs.get("class_expand", DEFAULT_CONFIG["class_expand"])
        self.use_last_conv = kwargs.get(
            "use_last_conv", DEFAULT_CONFIG["use_last_conv"]
        )
        self.act = kwargs.get("act", DEFAULT_CONFIG["act"])
        self.lr_mult_list = kwargs.get("lr_mult_list", DEFAULT_CONFIG["lr_mult_list"])
        self.net_config = kwargs.get("net_config", DEFAULT_CONFIG["net_config"])



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from paddlex import create_pipeline

pipeline = create_pipeline(pipeline="formula_recognition")

output = pipeline.predict(
    "./test_samples/general_formula_recognition01.png", use_layout_detection=True
)

# output = pipeline.predict(
#     "./test_samples/general_formula_recognition01.pdf",
#     use_layout_detection=True,
# )

# output = pipeline.predict(
#     "./test_samples/general_formula_recognition02.png",
#     use_layout_detection=False,
# )

# img_list = [ "./test_samples/general_formula_recognition03.png", \
#     "./test_samples/general_formula_recognition04.png", \
#         "./test_samples/general_formula_recognition05.png",]
# output = pipeline.predict(
#     img_list,
#     use_layout_detection=True,
# )

for res in output:
    print(res)
    res.print()
    res.save_to_img("./output")
    res.save_to_json("./output")



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import List

import numpy as np
import PIL
from PIL import Image, ImageDraw, ImageFont

from ....utils.deps import function_requires_deps, is_dep_available
from ....utils.fonts import PINGFANG_FONT
from ...common.result import BaseCVResult, JsonMixin
from ...utils.color_map import font_colormap, get_colormap

if is_dep_available("opencv-contrib-python"):
    import cv2


def draw_box(img: Image.Image, boxes: List[dict]) -> Image.Image:
    """
    Args:
        img (PIL.Image.Image): PIL image
        boxes (list): a list of dictionaries representing detection box information.
    Returns:
        img (PIL.Image.Image): visualized image
    """
    font_size = int(0.018 * int(img.width)) + 2
    font = ImageFont.truetype(PINGFANG_FONT.path, font_size, encoding="utf-8")

    draw_thickness = int(max(img.size) * 0.002)
    draw = ImageDraw.Draw(img)
    label2color = {}
    catid2fontcolor = {}
    color_list = get_colormap(rgb=True)

    for i, dt in enumerate(boxes):
        label, bbox, score = dt["label"], dt["coordinate"], dt["score"]
        if label not in label2color:
            color_index = i % len(color_list)
            label2color[label] = color_list[color_index]
            catid2fontcolor[label] = font_colormap(color_index)
        color = tuple(label2color[label])
        font_color = tuple(catid2fontcolor[label])

        if len(bbox) == 4:
            # draw bbox of normal object detection
            xmin, ymin, xmax, ymax = bbox
            rectangle = [
                (xmin, ymin),
                (xmin, ymax),
                (xmax, ymax),
                (xmax, ymin),
                (xmin, ymin),
            ]
        else:
            raise ValueError(
                f"Only support bbox format of [xmin,ymin,xmax,ymax] or [x1,y1,x2,y2,x3,y3,x4,y4], got bbox of shape {len(bbox)}."
            )

        # draw bbox
        draw.line(
            rectangle,
            width=draw_thickness,
            fill=color,
        )

        # draw label
        text = "{} {:.2f}".format(dt["label"], score)
        if tuple(map(int, PIL.__version__.split("."))) <= (10, 0, 0):
            tw, th = draw.textsize(text, font=font)
        else:
            left, top, right, bottom = draw.textbbox((0, 0), text, font)
            tw, th = right - left, bottom - top + 4
        if ymin < th:
            draw.rectangle([(xmin, ymin), (xmin + tw + 4, ymin + th + 1)], fill=color)
            draw.text((xmin + 2, ymin - 2), text, fill=font_color, font=font)
        else:
            draw.rectangle([(xmin, ymin - th), (xmin + tw + 4, ymin + 1)], fill=color)
            draw.text((xmin + 2, ymin - th - 2), text, fill=font_color, font=font)

        text_position = (bbox[2] + 2, bbox[1] - font_size // 2)
        if int(img.width) - bbox[2] < font_size:
            text_position = (
                int(bbox[2] - font_size * 1.1),
                bbox[1] - font_size // 2,
            )
        draw.text(text_position, str(i + 1), font=font, fill="red")

    return img


@function_requires_deps("opencv-contrib-python")
def restore_to_draw_masks(img_size, boxes):
    """
    Restores extracted masks to the original shape and draws them on a blank image.

    """
    restored_masks = []

    for i, box_info in enumerate(boxes):
        restored_mask = np.zeros(img_size, dtype=np.uint8)
        polygon = np.array(box_info["polygon_points"], dtype=np.int32)
        polygon = polygon.reshape((-1, 1, 2))  # shape: (N, 1, 2)
        cv2.fillPoly(restored_mask, [polygon], 1)
        restored_masks.append(restored_mask)

    return np.array(restored_masks)


def draw_mask(im, boxes, img_size):
    """
    Args:
        im (PIL.Image.Image): PIL image
        boxes (list): a list of dicts representing detection box information.
    Returns:
        img (PIL.Image.Image): visualized image
    """
    color_list = get_colormap(rgb=True)
    alpha = 0.5

    im = np.array(im).astype("float32")
    clsid2color = {}

    np_masks = restore_to_draw_masks(img_size, boxes)
    im_h, im_w = im.shape[:2]
    np_masks = np_masks[:, :im_h, :im_w]

    # draw mask
    for i, mask in enumerate(np_masks):
        clsid = int(boxes[i]["cls_id"])
        if clsid not in clsid2color:
            color_index = i % len(color_list)
            clsid2color[clsid] = np.array(color_list[color_index])
        color_mask = clsid2color[clsid]
        idx = np.nonzero(mask)
        im[idx[0], idx[1], :] = (1.0 - alpha) * im[
            idx[0], idx[1], :
        ] + alpha * color_mask

    img = Image.fromarray(np.uint8(im))
    font_size = int(0.018 * img.width) + 2
    font = ImageFont.truetype(PINGFANG_FONT.path, font_size, encoding="utf-8")
    draw = ImageDraw.Draw(img)
    label2color = {}
    catid2fontcolor = {}

    for i, box_info in enumerate(boxes):
        label = box_info["label"]
        score = box_info["score"]
        if label not in label2color:
            color_index = i % len(color_list)
            label2color[label] = color_list[color_index]
            catid2fontcolor[label] = font_colormap(color_index)
        color = tuple(label2color[label])
        font_color = tuple(catid2fontcolor[label])

        polygon_points = box_info["polygon_points"]

        image_left_top = (0, 0)
        image_right_top = (img.width, 0)
        left_top = min(
            polygon_points,
            key=lambda p: (
                (p[0] - image_left_top[0]) ** 2 + (p[1] - image_left_top[1]) ** 2
            ),
        )
        right_top = min(
            polygon_points,
            key=lambda p: (
                (p[0] - image_right_top[0]) ** 2 + (p[1] - image_right_top[1]) ** 2
            ),
        )

        # label
        text = "{} {:.2f}".format(label, score)
        if tuple(map(int, PIL.__version__.split("."))) <= (10, 0, 0):
            tw, th = draw.textsize(text, font=font)
        else:
            left, top, right, bottom = draw.textbbox((0, 0), text, font)
            tw, th = right - left, bottom - top + 4
        lx, ly = left_top
        if ly < th:
            draw.rectangle([(lx, ly), (lx + tw + 4, ly + th + 1)], fill=color)
            draw.text((lx + 2, ly - 2), text, fill=font_color, font=font)
        else:
            draw.rectangle([(lx, ly - th), (lx + tw + 4, ly + 1)], fill=color)
            draw.text((lx + 2, ly - th - 2), text, fill=font_color, font=font)

        # order
        order = box_info.get("order", None)
        if order:
            order_text = str(order)
            rx, ry = right_top
            text_position = (rx + 2, ry - font_size // 2)
            if int(img.width) - rx < font_size:
                text_position = (
                    int(rx - font_size * 1.1),
                    ry - font_size // 2,
                )
            draw.text(text_position, order_text, font=font, fill="red")

    return img


class LayoutAnalysisResult(BaseCVResult):

    def _to_img(self) -> Image.Image:
        """apply"""
        boxes = self["boxes"]
        image = Image.fromarray(self["input_img"][..., ::-1])
        ori_img_size = list(image.size)[::-1]
        if len(boxes) > 0 and "polygon_points" in boxes[0]:
            image = draw_mask(image, boxes, ori_img_size)
        else:
            image = draw_box(image, boxes)
        return {"res": image}

    def _to_str(self, *args, **kwargs):
        data = copy.deepcopy(self)
        data.pop("input_img")
        return JsonMixin._to_str(data, *args, **kwargs)

    def _to_json(self, *args, **kwargs):
        data = copy.deepcopy(self)
        data.pop("input_img")
        return JsonMixin._to_json(data, *args, **kwargs)



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from .... import UltraInferModel, ModelFormat
from .... import c_lib_wrap as C


class ResNet(UltraInferModel):
    def __init__(
        self,
        model_file,
        params_file="",
        runtime_option=None,
        model_format=ModelFormat.ONNX,
    ):
        """Load a image classification model exported by torchvision.ResNet.

        :param model_file: (str)Path of model file, e.g resnet/resnet50.onnx
        :param params_file: (str)Path of parameters file, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model, default is ONNX
        """

        # call super() to initialize the backend_option
        # the result of initialization will be saved in self._runtime_option
        super(ResNet, self).__init__(runtime_option)

        self._model = C.vision.classification.ResNet(
            model_file, params_file, self._runtime_option, model_format
        )
        # self.initialized shows the initialization of the model is successful or not

        assert self.initialized, "ResNet initialize failed."

    # Predict and return the inference result of "input_image".
    def predict(self, input_image, topk=1):
        """Classify an input image

        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :param topk: (int)The topk result by the classify confidence score, default 1
        :return: ClassifyResult
        """
        return self._model.predict(input_image, topk)

    # Implement the setter and getter method for variables
    @property
    def size(self):
        """
        Returns the preprocess image size, default size = [224, 224];
        """
        return self._model.size

    @property
    def mean_vals(self):
        """
        Returns the mean value of normalization, default mean_vals = [0.485f, 0.456f, 0.406f];
        """
        return self._model.mean_vals

    @property
    def std_vals(self):
        """
        Returns the std value of normalization, default std_vals = [0.229f, 0.224f, 0.225f];
        """
        return self._model.std_vals

    @size.setter
    def size(self, wh):
        assert isinstance(
            wh, (list, tuple)
        ), "The value to set `size` must be type of tuple or list."
        assert (
            len(wh) == 2
        ), "The value to set `size` must contains 2 elements means [width, height], but now it contains {} elements.".format(
            len(wh)
        )
        self._model.size = wh

    @mean_vals.setter
    def mean_vals(self, value):
        assert isinstance(
            value, list
        ), "The value to set `mean_vals` must be type of list."
        self._model.mean_vals = value

    @std_vals.setter
    def std_vals(self, value):
        assert isinstance(
            value, list
        ), "The value to set `std_vals` must be type of list."
        self._model.std_vals = value



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union

import paddle
import paddle.nn as nn
import paddle.nn.functional as F
from paddle import ParamAttr
from paddle.nn.initializer import KaimingNormal
from paddle.regularizer import L2Decay

from ....utils.benchmark import add_inference_operations, benchmark
from ...common.transformers.transformers import (
    BatchNormHFStateDictMixin,
    PretrainedModel,
)
from ._config_pp_ocrv5_mobile import PPOCRV5MobileDetConfig
from .pp_ocrv5_modules import DBHead, LearnableAffineBlock


def make_divisible(
    v: Union[int, float], divisor: int = 16, min_value: Optional[int] = None
) -> int:
    """
    make_divisible: Adjust channel number to be divisible by specified divisor (network width optimization)

    Args:
        v (Union[int, float]): Original channel number
        divisor (int, optional): Divisor for channel adjustment, default 16
        min_value (Optional[int], optional): Minimum channel number after adjustment, default None

    Returns:
        int: Adjusted channel number (integer)
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class Act(nn.Layer):
    """
    Act: Activation layer with Learnable Affine Block (LAB)

    Args:
        act (str): Activation type, "relu" or "hswish"
        lr_mult (float): Learning rate multiplier for LAB alpha
        lab_lr (float): Learning rate multiplier for LAB beta

    Returns:
        paddle.Tensor: Output tensor after activation and LAB
    """

    def __init__(self, act: str, lr_mult: float, lab_lr: float):
        super().__init__()
        if act == "hswish":
            self.act = nn.Hardswish()
        else:
            assert act == "relu"
            self.act = nn.ReLU()
        self.lab = LearnableAffineBlock(lr_mult=lr_mult, lab_lr=lab_lr)

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        return self.lab(self.act(x))


class ConvBNLayer(nn.Layer):
    """
    ConvBNLayer: Convolution + Batch Normalization combination layer

    Args:
        in_channels (int): Input channel number
        out_channels (int): Output channel number
        kernel_size (int): Convolution kernel size
        stride (int): Convolution stride
        lr_mult (float): Learning rate multiplier for conv/BN params
        groups (int, optional): Group convolution number, default 1

    Returns:
        paddle.Tensor: Output tensor after conv and BN
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        lr_mult: float,
        groups: int = 1,
    ):
        super().__init__()
        self.conv = nn.Conv2D(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=(kernel_size - 1) // 2,
            groups=groups,
            weight_attr=ParamAttr(initializer=KaimingNormal(), learning_rate=lr_mult),
            bias_attr=False,
        )

        self.bn = nn.BatchNorm2D(
            out_channels,
            weight_attr=ParamAttr(regularizer=L2Decay(0.0), learning_rate=lr_mult),
            bias_attr=ParamAttr(regularizer=L2Decay(0.0), learning_rate=lr_mult),
        )

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        x = self.conv(x)
        x = self.bn(x)
        return x


class LearnableRepLayer(nn.Layer):
    """
    LearnableRepLayer: Learnable representation layer with multi-branch convolution fusion

    Args:
        in_channels (int): Input channel number
        out_channels (int): Output channel number
        kernel_size (int): Convolution kernel size
        act (str): Activation type, "relu" or "hswish"
        stride (int): Convolution stride
        lr_mult (float): Learning rate multiplier for conv/BN/LAB params
        lab_lr (float): Learning rate multiplier for LAB beta
        num_conv_branches (int): Number of kxk conv branches
        groups (int, optional): Group convolution number, default 1

    Returns:
        paddle.Tensor: Output tensor after rep branches fusion and activation
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        act: str,
        stride: int,
        lr_mult: float,
        lab_lr: float,
        num_conv_branches: int,
        groups: int = 1,
    ):
        super().__init__()
        self.groups = groups
        self.stride = stride
        self.kernel_size = kernel_size
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_conv_branches = num_conv_branches
        self.padding = (kernel_size - 1) // 2

        self.identity = (
            nn.BatchNorm2D(
                num_features=in_channels,
                weight_attr=ParamAttr(learning_rate=lr_mult),
                bias_attr=ParamAttr(learning_rate=lr_mult),
            )
            if out_channels == in_channels and stride == 1
            else None
        )

        self.conv_kxk = nn.LayerList(
            [
                ConvBNLayer(
                    in_channels,
                    out_channels,
                    kernel_size,
                    stride,
                    groups=groups,
                    lr_mult=lr_mult,
                )
                for _ in range(self.num_conv_branches)
            ]
        )

        self.conv_1x1 = (
            ConvBNLayer(
                in_channels, out_channels, 1, stride, groups=groups, lr_mult=lr_mult
            )
            if kernel_size > 1
            else None
        )

        self.lab = LearnableAffineBlock(lr_mult=lr_mult, lab_lr=lab_lr)
        self.act = Act(act=act, lr_mult=lr_mult, lab_lr=lab_lr)

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:

        out = 0
        if self.identity is not None:
            out += self.identity(x)

        if self.conv_1x1 is not None:
            out += self.conv_1x1(x)

        for conv in self.conv_kxk:
            out += conv(x)

        out = self.lab(out)
        if self.stride != 2:
            out = self.act(out)
        return out


class SELayer(nn.Layer):
    """
    SELayer: Squeeze-and-Excitation channel attention layer

    Args:
        channel (int): Input/output channel number
        reduction (int): Channel reduction ratio for excitation
        lr_mult (float): Learning rate multiplier for conv params

    Returns:
        paddle.Tensor: Output tensor after channel attention weighting
    """

    def __init__(self, channel: int, reduction: int, lr_mult: float):
        super().__init__()
        if "npu" in paddle.device.get_device():
            self.avg_pool = nn.MeanPool2D(1, 1)
        else:
            self.avg_pool = nn.AdaptiveAvgPool2D(1)
        self.conv1 = nn.Conv2D(
            in_channels=channel,
            out_channels=channel // reduction,
            kernel_size=1,
            stride=1,
            padding=0,
            weight_attr=ParamAttr(learning_rate=lr_mult),
            bias_attr=ParamAttr(learning_rate=lr_mult),
        )
        self.relu = nn.ReLU()
        self.conv2 = nn.Conv2D(
            in_channels=channel // reduction,
            out_channels=channel,
            kernel_size=1,
            stride=1,
            padding=0,
            weight_attr=ParamAttr(learning_rate=lr_mult),
            bias_attr=ParamAttr(learning_rate=lr_mult),
        )
        self.hardsigmoid = nn.Hardsigmoid()

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        identity = x
        x = self.avg_pool(x)
        x = self.conv1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.hardsigmoid(x)
        x = paddle.multiply(x=identity, y=x)
        return x


class LCNetV3Block(nn.Layer):
    """
    LCNetV3Block: Depthwise separable convolution block with SE attention (LCNetV3)

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        act (str): Activation function type
        stride (int): Convolution stride for depthwise conv
        dw_size (int): Kernel size of depthwise convolution
        use_se (bool): Whether to use SE channel attention module
        conv_kxk_num (int): Number of conv branches in LearnableRepLayer
        reduction (int): Channel reduction ratio for SE module
        lr_mult (float): Learning rate multiplier for convolution parameters
        lab_lr (float): Learning rate multiplier for learnable representation layer

    Returns:
        paddle.Tensor: Output tensor after depthwise conv, SE (optional) and pointwise conv
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        act: str,
        stride: int,
        dw_size: int,
        use_se: bool,
        conv_kxk_num: int,
        reduction: int,
        lr_mult: float,
        lab_lr: float,
    ):
        super().__init__()
        self.use_se = use_se
        self.dw_conv = LearnableRepLayer(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=dw_size,
            act=act,
            stride=stride,
            groups=in_channels,
            num_conv_branches=conv_kxk_num,
            lr_mult=lr_mult,
            lab_lr=lab_lr,
        )
        if use_se:
            self.se = SELayer(in_channels, reduction=reduction, lr_mult=lr_mult)
        self.pw_conv = LearnableRepLayer(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=1,
            act=act,
            stride=1,
            num_conv_branches=conv_kxk_num,
            lr_mult=lr_mult,
            lab_lr=lab_lr,
        )

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        x = self.dw_conv(x)
        if self.use_se:
            x = self.se(x)
        x = self.pw_conv(x)
        return x


class PPLCNetV3(nn.Layer):
    """
    PPLCNetV3: Lightweight convolutional network with learnable representation layers

    Args:
        scale (float): Channel scale factor for network width adjustment
        conv_kxk_num (int): Number of conv branches in LearnableRepLayer of LCNetV3Block
        reduction (int): Channel reduction ratio for SE module in LCNetV3Block
        act (str): Activation function type used in LCNetV3Block
        lr_mult_list (List[float]): Learning rate multipliers for different layers, length must be 6
        lab_lr (float): Learning rate multiplier for learnable representation layer in LCNetV3Block
        net_config (Dict[str, Any]): Network configuration dict containing block parameters and output channels
        out_channels (int): Base number of output channels before scale adjustment
        **kwargs: Additional keyword arguments

    Returns:
        List[paddle.Tensor]: List of 4 feature tensors from different stages after 1x1 conv projection
    """

    def __init__(
        self,
        scale: float,
        conv_kxk_num: int,
        reduction: int,
        act: str,
        lr_mult_list: List[float],
        lab_lr: float,
        net_config: Dict[str, Any],
        out_channels: int,
        **kwargs,
    ):
        super().__init__()
        self.scale = scale
        self.lr_mult_list = lr_mult_list

        self.net_config = net_config
        self.out_channels = make_divisible(out_channels * scale)

        assert isinstance(
            self.lr_mult_list, (list, tuple)
        ), "lr_mult_list should be in (list, tuple) but got {}".format(
            type(self.lr_mult_list)
        )
        assert (
            len(self.lr_mult_list) == 6
        ), "lr_mult_list length should be 6 but got {}".format(len(self.lr_mult_list))

        self.conv1 = ConvBNLayer(
            in_channels=3,
            out_channels=make_divisible(16 * scale),
            kernel_size=3,
            stride=2,
            lr_mult=self.lr_mult_list[0],
        )

        def _build_blocks(block_key, lr_mult_idx):
            return nn.Sequential(
                *[
                    LCNetV3Block(
                        in_channels=make_divisible(in_c * scale),
                        out_channels=make_divisible(out_c * scale),
                        act=act,
                        dw_size=k,
                        stride=s,
                        use_se=se,
                        conv_kxk_num=conv_kxk_num,
                        reduction=reduction,
                        lr_mult=self.lr_mult_list[lr_mult_idx],
                        lab_lr=lab_lr,
                    )
                    for i, (k, in_c, out_c, s, se) in enumerate(
                        self.net_config[block_key]
                    )
                ]
            )

        self.blocks2 = _build_blocks("blocks2", 1)
        self.blocks3 = _build_blocks("blocks3", 2)
        self.blocks4 = _build_blocks("blocks4", 3)
        self.blocks5 = _build_blocks("blocks5", 4)
        self.blocks6 = _build_blocks("blocks6", 5)

        mv_c = self.net_config["layer_list_out_channels"]

        self.out_channels = [
            make_divisible(self.net_config["blocks3"][-1][2] * scale),
            make_divisible(self.net_config["blocks4"][-1][2] * scale),
            make_divisible(self.net_config["blocks5"][-1][2] * scale),
            make_divisible(self.net_config["blocks6"][-1][2] * scale),
        ]

        self.layer_list = nn.LayerList(
            [
                nn.Conv2D(self.out_channels[0], int(mv_c[0] * scale), 1, 1, 0),
                nn.Conv2D(self.out_channels[1], int(mv_c[1] * scale), 1, 1, 0),
                nn.Conv2D(self.out_channels[2], int(mv_c[2] * scale), 1, 1, 0),
                nn.Conv2D(self.out_channels[3], int(mv_c[3] * scale), 1, 1, 0),
            ]
        )
        self.out_channels = [
            int(mv_c[0] * scale),
            int(mv_c[1] * scale),
            int(mv_c[2] * scale),
            int(mv_c[3] * scale),
        ]

    def forward(self, x: paddle.Tensor) -> List[paddle.Tensor]:
        out_list = []
        x = self.conv1(x)

        x = self.blocks2(x)
        x = self.blocks3(x)
        out_list.append(x)
        x = self.blocks4(x)
        out_list.append(x)
        x = self.blocks5(x)
        out_list.append(x)
        x = self.blocks6(x)
        out_list.append(x)

        out_list[0] = self.layer_list[0](out_list[0])
        out_list[1] = self.layer_list[1](out_list[1])
        out_list[2] = self.layer_list[2](out_list[2])
        out_list[3] = self.layer_list[3](out_list[3])
        return out_list


class SEModule(nn.Layer):
    """
    SEModule: Simplified Squeeze-and-Excitation channel attention module

    Args:
        in_channels (int): Number of input channels
        reduction (int): Channel reduction ratio for excitation layer

    Returns:
        paddle.Tensor: Output tensor after channel attention weighting
    """

    def __init__(self, in_channels: int, reduction: int):
        super(SEModule, self).__init__()
        if "npu" in paddle.device.get_device():
            self.avg_pool = nn.MeanPool2D(1, 1)
        else:
            self.avg_pool = nn.AdaptiveAvgPool2D(1)
        self.conv1 = nn.Conv2D(
            in_channels=in_channels,
            out_channels=in_channels // reduction,
            kernel_size=1,
            stride=1,
            padding=0,
        )
        self.conv2 = nn.Conv2D(
            in_channels=in_channels // reduction,
            out_channels=in_channels,
            kernel_size=1,
            stride=1,
            padding=0,
        )

    def forward(self, inputs: paddle.Tensor) -> paddle.Tensor:
        outputs = self.avg_pool(inputs)
        outputs = self.conv1(outputs)
        outputs = F.relu(outputs)
        outputs = self.conv2(outputs)
        outputs = F.hardsigmoid(outputs, slope=0.2, offset=0.5)
        return inputs * outputs


class RSELayer(nn.Layer):
    """
    RSELayer: Residual SE layer with convolution and shortcut connection

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int): Kernel size of convolution layer
        shortcut (bool): Whether to add shortcut connection (residual) with SE output
        reduction (int): Channel reduction ratio for SE module

    Returns:
        paddle.Tensor: Output tensor after convolution, SE attention and optional shortcut
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        shortcut: bool,
        reduction: int,
    ):
        super(RSELayer, self).__init__()
        weight_attr = paddle.nn.initializer.KaimingUniform()
        self.out_channels = out_channels
        self.in_conv = nn.Conv2D(
            in_channels=in_channels,
            out_channels=self.out_channels,
            kernel_size=kernel_size,
            padding=int(kernel_size // 2),
            weight_attr=ParamAttr(initializer=weight_attr),
            bias_attr=False,
        )
        self.se_block = SEModule(self.out_channels, reduction=reduction)
        self.shortcut = shortcut

    def forward(self, ins: paddle.Tensor) -> paddle.Tensor:
        x = self.in_conv(ins)
        if self.shortcut:
            out = x + self.se_block(x)
        else:
            out = self.se_block(x)
        return out


class RSEFPN(nn.Layer):
    """
    RSEFPN: Feature Pyramid Network with Residual SE attention

    Args:
        in_channels (List[int]): List of input channel numbers for multi-scale feature maps
        out_channels (int): Number of output channels for RSELayer convolution
        shortcut (bool): Whether to use shortcut connection in RSELayer
        reduction (int): Channel reduction ratio for SE module in RSELayer
        **kwargs: Additional keyword arguments

    Returns:
        paddle.Tensor: Fused feature tensor after multi-scale feature aggregation and concatenation
    """

    def __init__(
        self,
        in_channels: List[int],
        out_channels: int,
        shortcut: bool,
        reduction: int,
        **kwargs,
    ):
        super(RSEFPN, self).__init__()
        self.out_channels = out_channels
        self.ins_conv = nn.LayerList()
        self.inp_conv = nn.LayerList()

        for i in range(len(in_channels)):
            self.ins_conv.append(
                RSELayer(
                    in_channels[i],
                    out_channels,
                    kernel_size=1,
                    shortcut=shortcut,
                    reduction=reduction,
                )
            )
            self.inp_conv.append(
                RSELayer(
                    out_channels,
                    out_channels // 4,
                    kernel_size=3,
                    shortcut=shortcut,
                    reduction=reduction,
                )
            )

    def forward(self, x: List[paddle.Tensor]) -> paddle.Tensor:
        c2, c3, c4, c5 = x

        in5 = self.ins_conv[3](c5)
        in4 = self.ins_conv[2](c4)
        in3 = self.ins_conv[1](c3)
        in2 = self.ins_conv[0](c2)

        out4 = in4 + F.upsample(
            in5, scale_factor=2, mode="nearest", align_mode=1
        )  # 1/16
        out3 = in3 + F.upsample(
            out4, scale_factor=2, mode="nearest", align_mode=1
        )  # 1/8
        out2 = in2 + F.upsample(
            out3, scale_factor=2, mode="nearest", align_mode=1
        )  # 1/4

        p5 = self.inp_conv[3](in5)
        p4 = self.inp_conv[2](out4)
        p3 = self.inp_conv[1](out3)
        p2 = self.inp_conv[0](out2)

        p5 = F.upsample(p5, scale_factor=8, mode="nearest", align_mode=1)
        p4 = F.upsample(p4, scale_factor=4, mode="nearest", align_mode=1)
        p3 = F.upsample(p3, scale_factor=2, mode="nearest", align_mode=1)

        fuse = paddle.concat([p5, p4, p3, p2], axis=1)
        return fuse


class PPOCRV5MobileDet(BatchNormHFStateDictMixin, PretrainedModel):
    """
    PPOCRV5MobileDet: Lightweight OCR detection model based on PPLCNetV3, RSEFPN and DBHead

    Args:
        config (PPOCRV5MobileDetConfig): Configuration object containing model hyperparameters

    Returns:
        List: List containing the detection output tensor (converted to numpy array on CPU)
    """

    config_class = PPOCRV5MobileDetConfig

    def __init__(self, config: PPOCRV5MobileDetConfig):
        super().__init__(config)

        self.backbone_scale = config.backbone_scale
        self.backbone_det = config.backbone_det
        self.backbone_conv_kxk_num = config.backbone_conv_kxk_num
        self.backbone_reduction = config.backbone_reduction
        self.backbone_act = config.backbone_act
        self.backbone_lr_mult_list = config.backbone_lr_mult_list
        self.backbone_lab_lr = config.backbone_lab_lr
        self.backbone_net_config = config.backbone_net_config
        self.backbone_out_channels = config.backbone_out_channels

        self.neck_out_channels = config.neck_out_channels
        self.neck_shortcut = config.neck_shortcut

        self.head_k = config.head_k
        self.head_kernel_list = config.head_kernel_list
        self.head_fix_nan = config.head_fix_nan

        self.backbone = PPLCNetV3(
            scale=self.backbone_scale,
            conv_kxk_num=self.backbone_conv_kxk_num,
            reduction=self.backbone_reduction,
            act=self.backbone_act,
            lr_mult_list=self.backbone_lr_mult_list,
            lab_lr=self.backbone_lab_lr,
            net_config=self.backbone_net_config,
            out_channels=self.backbone_out_channels,
        )

        neck_in_channels = self.backbone.out_channels
        self.neck = RSEFPN(
            in_channels=neck_in_channels,
            out_channels=self.neck_out_channels,
            shortcut=self.neck_shortcut,
            reduction=self.backbone_reduction,
        )

        head_in_channels = self.neck_out_channels
        self.head = DBHead(
            in_channels=head_in_channels,
            k=self.head_k,
            kernel_list=self.head_kernel_list,
            fix_nan=self.head_fix_nan,
        )

    add_inference_operations("pp_ocrv5_mobile_det_forward")

    @benchmark.timeit_with_options(name="pp_ocrv5_mobile_det_forward")
    def forward(self, x: List) -> List:

        if isinstance(x, (list, tuple)):
            x = x[0]
        if not isinstance(x, paddle.Tensor):
            x = paddle.to_tensor(x)

        x = self.backbone(x)
        x = self.neck(x)
        x = self.head(x)

        return [x.cpu().numpy()]



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from pathlib import Path

from ...common.result import BaseCVResult, LatexMixin, MarkdownMixin, WordMixin


class MarkdownResult(BaseCVResult, MarkdownMixin, WordMixin, LatexMixin):
    def __init__(self, data) -> None:
        """Initializes a new instance of the class with the specified data."""
        super().__init__(data)
        MarkdownMixin.__init__(self)

    def _get_input_fn(self):
        fn = super()._get_input_fn()
        if (page_idx := self.get("page_index", None)) is not None:
            fp = Path(fn)
            stem, suffix = fp.stem, fp.suffix
            fn = f"{stem}_{page_idx}{suffix}"
        if (language := self.get("language", None)) is not None:
            fp = Path(fn)
            stem, suffix = fp.stem, fp.suffix
            fn = f"{stem}_{language}{suffix}"
        return fn

    def _to_markdown(self, pretty=True, show_formula_number=False) -> dict:
        return self

    # in order to make MarkdownResult support save_to_word
    def _to_word(self, save_path) -> dict:
        from bs4 import BeautifulSoup

        md_text = self.get("markdown_texts", "")

        def set_paragraph_style(
            paragraph, bold=False, align="left", font_size=11, color=None
        ):
            from docx.enum.text import WD_PARAGRAPH_ALIGNMENT
            from docx.oxml.ns import qn
            from docx.shared import Pt, RGBColor

            # Set paragraph style uniformly
            run = paragraph.runs[0] if paragraph.runs else paragraph.add_run()
            run.font.name = "Times New Roman"
            run._element.rPr.rFonts.set(qn("w:eastAsia"), "宋体")
            run.font.size = Pt(font_size)
            run.bold = bold
            if color:
                run.font.color.rgb = RGBColor(*color)
            if align == "center":
                paragraph.alignment = WD_PARAGRAPH_ALIGNMENT.CENTER
            elif align == "right":
                paragraph.alignment = WD_PARAGRAPH_ALIGNMENT.RIGHT
            else:
                paragraph.alignment = WD_PARAGRAPH_ALIGNMENT.LEFT

        def add_image(paragraph, src, width_percent):
            from docx.enum.text import WD_PARAGRAPH_ALIGNMENT
            from docx.shared import Inches

            if os.path.exists(src):
                try:
                    width_in_inches = Inches(width_percent / 100 * 6.0)
                    run = paragraph.add_run()
                    run.add_picture(src, width=width_in_inches)
                    paragraph.alignment = WD_PARAGRAPH_ALIGNMENT.CENTER
                except Exception as e:
                    paragraph.add_run(f"[fail load image: {src}]")
            else:
                paragraph.add_run(f"[image not exist: {src}]")

        def add_table(document, table_html):
            """
            Parsing HTML table and add to Word
            """
            soup = BeautifulSoup(table_html, "html.parser")
            table_tag = soup.find("table")
            if not table_tag:
                return

            rows = table_tag.find_all("tr")
            if not rows:
                return

            # Calculate the maximum number of columns to avoid out-of-bounds errors
            max_cols = max(len(row.find_all(["td", "th"])) for row in rows)
            table = document.add_table(rows=len(rows), cols=max_cols)
            table.style = "Table Grid"

            for i, row in enumerate(rows):
                cells = row.find_all(["td", "th"])
                for j in range(max_cols):
                    if j < len(cells):
                        text = cells[j].get_text(strip=True)
                        table.cell(i, j).text = text
                    else:
                        table.cell(i, j).text = ""

        def process_md_page(document, md_text, output_path):

            # Process single page conten
            lines = md_text.strip().split("\n")
            for line in lines:
                line = line.strip()
                if not line:
                    continue

                title_color = (0, 0, 255)
                if line.startswith("##### "):
                    p = document.add_paragraph(line[6:])
                    set_paragraph_style(p, bold=True, font_size=10)
                elif line.startswith("#### "):
                    p = document.add_paragraph(line[5:])
                    set_paragraph_style(p, bold=True, font_size=11)
                elif line.startswith("### "):
                    p = document.add_paragraph(line[4:])
                    set_paragraph_style(p, bold=True, font_size=12)
                elif line.startswith("## "):
                    p = document.add_paragraph(line[3:])
                    set_paragraph_style(p, bold=True, font_size=14)
                elif line.startswith("# "):
                    p = document.add_paragraph(line[2:])
                    set_paragraph_style(p, bold=True, font_size=16)

                # Handle centered content
                elif line.startswith("<div") and "text-align: center" in line:
                    soup = BeautifulSoup(line, "html.parser")
                    div = soup.find("div")
                    if not div:
                        continue
                    if div.img:
                        img = div.img
                        src = img.get("src")
                        width_attr = img.get("width", "100%").replace("%", "")
                        width_percent = float(width_attr) if width_attr else 100
                        p = document.add_paragraph()
                        add_image(p, f"{output_path}/{src}", width_percent)
                    elif div.table:
                        add_table(document, str(div))
                    else:
                        text = div.get_text(strip=True)
                        if text:
                            p = document.add_paragraph(text)
                            set_paragraph_style(
                                p, bold=True, align="center", color=title_color
                            )

                # Handle HTML tables
                elif "<table" in line:
                    add_table(document, line)

                # Normal paragraph
                else:
                    p = document.add_paragraph(line)
                    set_paragraph_style(p, font_size=11)

        from docx import Document

        document = Document()
        process_md_page(document, md_text, save_path)

        return document

    # in order to make MarkdownResult support save_to_latex
    def _to_latex(self, save_path) -> str:
        from bs4 import BeautifulSoup

        def escape_latex_outside_formula(s: str) -> str:
            """
            Escape LaTeX special characters while preserving formulas.
            """
            if not s:
                return ""

            placeholders = []

            def repl(m):
                placeholders.append(m.group(0))
                return f"@@FORMULA{len(placeholders)-1}@@"

            # Extract formulas
            formula_pat = re.compile(
                r"(\$\$.*?\$\$|\$.*?\$|\\\[.*?\\\]|\\\(.*?\\\))", re.DOTALL
            )
            tmp = formula_pat.sub(repl, s)

            tmp = (
                tmp.replace("\\", "\\textbackslash{}")
                .replace("&", "\\&")
                .replace("%", "\\%")
                .replace("$", "\\$")
                .replace("#", "\\#")
                .replace("_", "\\_")
                .replace("{", "\\{")
                .replace("}", "\\}")
                .replace("~", "\\textasciitilde{}")
                .replace("^", "\\textasciicircum{}")
            )

            # Restore formulas
            for i, f in enumerate(placeholders):
                tmp = tmp.replace(f"@@FORMULA{i}@@", f)
            return tmp

        #
        def get_image_width_from_md_line(line, default_ratio=0.8):
            """
            Parse the image width attribute.
            """
            m = re.search(r'width\s*=\s*["\']?(\d+)%?["\']?', line)
            if m:
                val = int(m.group(1))
                return max(0.01, min(val / 100.0, 1.0))
            m2 = re.search(r"width\s*:\s*(\d+)%", line)
            if m2:
                val = int(m2.group(1))
                return max(0.01, min(val / 100.0, 1.0))
            return default_ratio

        def process_table_html(content) -> str:
            """
            Process table content.
            """
            if "<table" in content:
                soup = BeautifulSoup(content, "html.parser")
                rows = []
                for tr in soup.find_all("tr"):
                    row = []
                    for td in tr.find_all(["td", "th"]):
                        text = td.get_text(strip=True)
                        row.append(escape_latex_outside_formula(text))
                    rows.append(row)
            else:
                rows = [
                    [escape_latex_outside_formula(c) for c in row.split("\t")]
                    for row in content.splitlines()
                    if row.strip()
                ]

            if not rows:
                return ""

            col_count = max(len(r) for r in rows)
            norm_rows = [r + [""] * (col_count - len(r)) for r in rows]
            col_format = " ".join(
                [">{\\raggedright\\arraybackslash}X" for _ in range(col_count)]
            )

            latex = "\\begin{center}\n\\renewcommand{\\arraystretch}{1.5}\n"
            latex += f"\\begin{{tabularx}}{{\\textwidth}}{{{col_format}}}\n\\toprule\n"
            for i, row in enumerate(norm_rows):
                latex += " & ".join(row) + " \\\\\n"
                if i == 0:
                    latex += "\\midrule\n"
            latex += "\\bottomrule\n\\end{tabularx}\n\\end{center}\n\n"
            return latex

        def process_paragraph(s: str) -> str:
            """
            Process text paragraphs, preserving formulas.
            """
            paragraphs = re.split(r"\n\s*\n", s)
            processed_paras = []
            for p in paragraphs:
                p = p.strip()
                if not p:
                    continue
                processed_paras.append("\\par " + escape_latex_outside_formula(p))
            return "\n\n".join(processed_paras) + "\n\n"

        def process_md_line(line: str, save_path) -> str:
            """
            Process a single line.
            """
            line = line.strip()
            if not line:
                return ""

            if line.startswith("##### "):
                return f"\\paragraph*{{{escape_latex_outside_formula(line[6:].strip())}}}\n\n"
            if line.startswith("#### "):
                return f"\\subsubsection*{{{escape_latex_outside_formula(line[5:].strip())}}}\n\n"
            if line.startswith("### "):
                return f"\\subsection*{{{escape_latex_outside_formula(line[4:].strip())}}}\n\n"
            if line.startswith("## "):
                return f"\\section*{{{escape_latex_outside_formula(line[3:].strip())}}}\n\n"
            if line.startswith("# "):
                return f"\\section*{{{escape_latex_outside_formula(line[2:].strip())}}}\n\n"

            if "<div" in line and "text-align: center" in line:
                soup = BeautifulSoup(line, "html.parser")
                div = soup.find("div")
                if div:
                    if div.img:
                        img = div.img
                        src = img.get("src")
                        src = f"{save_path}/{src}"
                        width_ratio = get_image_width_from_md_line(str(img))
                        return (
                            f"\\begin{{figure}}[h]\n\\centering\n"
                            f"\\includegraphics[width={width_ratio:.2f}\\linewidth]{{{src}}}\n"
                            f"\\end{{figure}}\n\n"
                        )
                    if div.table:
                        return process_table_html(str(div))
                    text = div.get_text(strip=True)
                    if text:
                        return f"\\begin{{center}}{escape_latex_outside_formula(text)}\\end{{center}}\n\n"

            if "<table" in line:
                return process_table_html(line)
            return process_paragraph(line)

        latex_lines = [
            "\\documentclass[12pt]{article}",
            "\\usepackage{xeCJK}",
            "\\usepackage{fontspec}",
            "\\usepackage{graphicx}",
            "\\usepackage{amsmath}",
            "\\usepackage{geometry}",
            "\\usepackage{fancyhdr}",
            "\\usepackage{indentfirst}",
            "\\usepackage{caption}",
            "\\usepackage{tabularx, booktabs}",
            "\\usepackage{amssymb}",
            "\\usepackage{amsfonts}",
            "\\geometry{a4paper, margin=1in}",
            "\\setCJKmainfont{Droid Sans Fallback}",
            "\\setmainfont{DejaVu Serif}",
            "\\setsansfont{Lato}",
            "\\setmonofont{Latin Modern Mono}",
            "\\pagestyle{fancy}",
            "\\setlength{\\parindent}{2em}",
            "\\begin{document}\n",
        ]

        md_text = self.get("markdown_texts", "")
        for line in md_text.splitlines():
            latex_lines.append(process_md_line(line, save_path))

        latex_lines.append("\\end{document}")

        return "\n".join(latex_lines)


class DocumentResult(BaseCVResult, WordMixin):
    def __init__(self, data) -> None:
        """
        Initializes a new instance of the class with the specified data.
        """
        super().__init__(data)
        WordMixin.__init__(self)

    def _get_input_fn(self):
        fn = super()._get_input_fn()
        if (page_idx := self.get("page_index", None)) is not None:
            fp = Path(fn)
            stem, suffix = fp.stem, fp.suffix
            fn = f"{stem}_{page_idx}{suffix}"
        if (language := self.get("language", None)) is not None:
            fp = Path(fn)
            stem, suffix = fp.stem, fp.suffix
            fn = f"{stem}_{language}{suffix}"
        return fn

    def _to_word(self) -> dict:
        return self


class LatexResult(BaseCVResult, LatexMixin):
    def __init__(self, data) -> None:
        super().__init__(data)
        LatexMixin.__init__(self)

    def _get_input_fn(self):
        fn = super()._get_input_fn()
        if (page_idx := self.get("page_index", None)) is not None:
            fp = Path(fn)
            stem, suffix = fp.stem, fp.suffix
            fn = f"{stem}_{page_idx}{suffix}"
        if (language := self.get("language", None)) is not None:
            fp = Path(fn)
            stem, suffix = fp.stem, fp.suffix
            fn = f"{stem}_{language}{suffix}"
        return fn

    def _to_latex(self) -> dict:
        return self



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from paddlex import create_pipeline

pipeline = create_pipeline(pipeline="doc_preprocessor")

output = pipeline.predict(
    "./test_samples/img_rot180_demo.jpg",
    use_doc_orientation_classify=True,
    use_doc_unwarping=False,
)

# output = pipeline.predict(
#     "./test_samples/img_rot180_demo.jpg",
#     use_doc_orientation_classify=False,
#     use_doc_unwarping=True,
# )

# output = pipeline.predict(
#     "./test_samples/img_rot180_demo.jpg",
#     use_doc_orientation_classify=True,
#     use_doc_unwarping=True,
# )

# output = pipeline.predict(
#     "./test_samples/img_rot180_demo.jpg",
#     use_doc_orientation_classify=False,
#     use_doc_unwarping=False,
# )

# output = pipeline.predict(
#     "./test_samples/doc_distort_test.jpg",
#     use_doc_orientation_classify=True,
#     use_doc_unwarping=True
# )

for res in output:
    print(res)
    res.print()
    res.save_to_img("./output")
    res.save_to_json("./output")
